from typing import Optional

from anyio.to_thread import run_sync

from ice.agents.base import Agent
from ice.nn.tfew import load_inference_model
from ice.nn.tfew import PromptedClassificationInput
from ice.nn.tfew import PromptedClassificationOutput
from ice.utils import DynamicBatcher


class TFew(Agent):
    def __init__(
        self,
        origin_model_name: str,
        lora_weights_path: str,
        max_batch_size: int = 32,
        batch_interval_seconds=0.1,
    ):
        self.model = load_inference_model(
            origin_model_name=origin_model_name, lora_weights_path=lora_weights_path
        )
        self.batcher = DynamicBatcher(
            handler=self.model.predict,
            max_batch_size=max_batch_size,
            batch_interval_seconds=batch_interval_seconds,
        )

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: Optional[str] = None,
        verbose: bool = False
    ) -> tuple[dict[str, float], Optional[str]]:
        inp = PromptedClassificationInput(prompt=prompt, choices=choices)

        def run_batch() -> PromptedClassificationOutput:
            return self.batcher.process(arg=inp)

        output = await run_sync(run_batch)

        # FIXME: Return full dict of probabilities
        return {output.prediction: output.prob}, None



import itertools

from collections.abc import Sequence
from typing import Optional

from pydantic import BaseModel
from tqdm import tqdm

from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import get_gold_standards
from ice.metrics.gold_standards import GoldStandard
from ice.metrics.gold_standards import GoldStandardSplit
from ice.paper import Paper
from ice.recipes.consort_flow.types import ConsortFlow
from ice.recipes.consort_flow.types import SampleSize
from ice.recipes.program_search.nodes.decontext.decontextualize import paper_decontext
from ice.recipes.program_search.nodes.select.dynamic import best_negative_example
from ice.recipes.program_search.nodes.select.dynamic import first_positive_example
from ice.recipes.program_search.nodes.select.dynamic import make_examples
from ice.recipes.program_search.nodes.select.dynamic import SelectionExample
from ice.recipes.program_search.nodes.select.prompts import render_selection_example
from ice.recipes.program_search.nodes.select.prompts import RenderableSelectionExample
from ice.recipes.program_search.types import Selection
from ice.recipes.program_search.types import sentences
from ice.settings import settings
from ice.utils import map_async


def get_consort_gs(document_id: str) -> Optional[GoldStandard[ConsortFlow]]:
    return get_gold_standard(
        document_id=document_id,
        question_short_name="consort_flow",
        model_type=ConsortFlow,
    )


def consort_gs_split(
    split: GoldStandardSplit, question_short_name: str
) -> Sequence[GoldStandard[ConsortFlow]]:
    golds = get_gold_standards(
        question_short_name=question_short_name, model_type=ConsortFlow
    )
    return [gs for gs in golds if gs.split == split]


def paper_to_allocation_gold_standards(
    paper: Paper,
) -> Sequence[tuple[str, Sequence[Selection], Sequence[str]]]:
    gs = get_consort_gs(paper.document_id)
    texts = sentences(paper)
    if not gs or not gs.parsed_answer:
        return []

    return [
        (
            f"The {exp.name} experiment included {len(exp.arms or [])} arms: {', '.join((arm.name for arm in exp.arms or []))}. How many participants were initially allocated to the {arm.name} arm of the {exp.name} experiment?",
            texts,
            arm.allocated.quotes
            if arm.allocated and isinstance(arm.allocated, SampleSize)
            else [],
        )
        for exp in gs.parsed_answer.experiments
        for arm in (exp.arms or [])
    ]


# def paper_to_experiments_gold_standard(paper: Paper) -> Sequence[tuple[str, Sequence[Selection], Sequence[str]]]:
#     gs = get_ea_gs(paper.document_id)
#     texts = sentences(paper)
#     if not gs or not gs.parsed_answer:
#         return []


class GoldStandardExample(BaseModel):
    question: str
    texts: Sequence[Selection]
    gs_quotes: Sequence[str]


def gold_standard_examples(
    papers: Sequence[Paper],
) -> Sequence[GoldStandardExample]:
    return [
        GoldStandardExample(question=question, texts=texts, gs_quotes=gs_quotes)
        for result in tqdm(
            map(paper_to_allocation_gold_standards, papers),
            desc="Creating gold standard examples",
            total=len(papers),
        )
        for question, texts, gs_quotes in result
    ]


def download_papers(
    split: str = "validation", question_short_name: str = "consort_flow"
):
    paper_dir = settings.PAPER_DIR
    doc_ids = {p.document_id for p in consort_gs_split(split, question_short_name)}  # type: ignore[arg-type]
    paper_files = [f for f in paper_dir.iterdir() if f.name in doc_ids]
    return [
        Paper.load(f)
        for f in tqdm(paper_files, desc="Loading papers for gold standards")
    ]


async def gold_standard_examples_except(
    document_ids_to_exclude: Sequence[str],
    require_quotes: bool = True,
    limit: Optional[int] = None,
    decontextualize: bool = False,
) -> Sequence[GoldStandardExample]:
    blocklist_ids = set(document_ids_to_exclude)
    papers = [p for p in download_papers() if p.document_id not in blocklist_ids]
    papers = papers[: (limit or len(papers))]
    if decontextualize:
        papers = [(await paper_decontext(p)) for p in papers]
    gses = gold_standard_examples(papers)
    return [gse for gse in gses if gse.gs_quotes or not require_quotes]


async def select_examples(
    examples: Sequence[GoldStandardExample], *, n: int, step: int, max_existing: int
) -> Sequence[tuple[str, SelectionExample]]:
    async def make_examples_for_paper(
        paper: GoldStandardExample,
    ) -> Sequence[tuple[str, SelectionExample]]:
        all_examples = await make_examples(
            texts=paper.texts,
            gs_quotes=paper.gs_quotes,
            n=n,
            step=step,
            max_existing=max_existing,
        )
        first_positive = first_positive_example(all_examples)
        best_negative = best_negative_example(all_examples)
        found_examples = filter(None, (first_positive, best_negative))
        return [(paper.question, found_example) for found_example in found_examples]

    all_examples = await map_async(examples, make_examples_for_paper)
    return list(itertools.chain.from_iterable(all_examples))


def narrow_examples(
    examples: Sequence[tuple[str, SelectionExample]], max_examples: int
):
    # TODO: Get much fancier about selectiong examples that make good prompts
    example_types: set[str] = set()
    examples_to_keep: list[tuple[str, SelectionExample]] = []
    for question, example in examples:
        if len(examples_to_keep) == max_examples:
            break
        example_type = (
            f"{example.selection[0].p.document_id}-{bool(example.positive_idxs)}"
        )
        if example_type in example_types:
            continue
        example_types.add(example_type)
        examples_to_keep.append((question, example))

    return examples_to_keep


async def selection_examples_for_paper(
    paper: Paper,
    max_examples: int = 5,
    limit_papers: Optional[int] = None,
    decontextualize: bool = False,
) -> list[RenderableSelectionExample]:
    golds = await gold_standard_examples_except(
        [paper.document_id], limit=limit_papers, decontextualize=decontextualize
    )
    examples = await select_examples(golds, n=5, step=5, max_existing=3)
    return [
        render_selection_example(question=question, example=example)
        for question, example in narrow_examples(examples, max_examples=max_examples)
    ]



"""
Papers that explicitly mention blinding of participants:
- awasthi-2013.pdf
- cheng-2020.pdf
- deutschmann-2019.pdf
- dicko-2008.pdf
- keenan-2018.pdf
- matangila-2015.pdf
- miguel-kremer-2004.pdf
- routledge-2006.pdf
- smithuis-2013.pdf

Papers that explicitly mention blinding of personnel:
- awasthi-2013.pdf
- bassi-2018.pdf (?)
- deutschmann-2019.pdf
- dicko-2008.pdf
- keenan-2018.pdf
- matangila-2015.pdf
- nyqvist-218.pdf
- routledge-2006.pdf
- vittengl-2009.pdf
"""
import itertools

from typing import Any
from typing import Literal
from typing import Optional
from typing import Union

import rich
import tqdm

from pydantic import BaseModel
from rich.panel import Panel
from thefuzz import fuzz

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import get_gold_standards
from ice.metrics.gold_standards import GoldStandard
from ice.metrics.gold_standards import list_experiments
from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import Recipe

# from rich.pretty import pprint

Group = Union[Literal["participants"], Literal["personnel"]]


def flatten(xs):
    return itertools.chain(*xs)


def quote(s: str):
    return f'"{s}"'


def sort_by_recall(gold_items: list[str], test_items: list[str]):
    test_string = " ".join(test_items)
    recalled_items = [
        gold_item
        for gold_item in gold_items
        if fuzz.partial_ratio(gold_item, test_string) > 95
    ]
    # for gold_item in gold_items:
    #     print(gold_item, fuzz.partial_ratio(gold_item, test_string))
    missed_items = [
        gold_item for gold_item in gold_items if gold_item not in recalled_items
    ]
    return recalled_items, missed_items


def get_blinding_gold_standards(
    *,
    paper: Paper,
    group: Group,
    explicitness: Union[Literal["explicit"], Literal["full"]] = "explicit",
    intervention: str,
) -> GoldStandard:
    question_short_name = f"blinding_{group}_{explicitness}"
    gold_standards = get_gold_standards(
        document_id=paper.document_id,
        question_short_name=question_short_name,
        experiment=intervention,
    )
    # if not gold_standards:
    #     fake_gold_standard = GoldStandard(
    #         document_id=paper.document_id,
    #         question_short_name=question_short_name,
    #         experiment=intervention,
    #         answer="",
    #         quotes=["bla"],
    #     )
    #     return fake_gold_standard
    assert len(gold_standards) == 1, gold_standards
    return gold_standards[0]


class InterventionResult(BaseModel):
    document_id: str
    question_short_name: str
    experiment: str
    answer: Optional[str]
    quotes: list[str]


class ParagraphResult(BaseModel):
    quotes: list[str]
    paragraph: Paragraph


PARTICIPANTS_CONTEXT = """Instructions for each paragraph:

What does the paragraph say explicitly about whether the participants (i.e. the people who received the intervention) were blinded to whether they were in the treatment or control group?

Include all relevant quotes:
- Everything the study explicitly says about whether participants were blinded. If the paper says that the experiment was "open label", that means participants knew whether they were in the treatment or control group. So, the experiment was not blinded.
- All efforts that the research team made to prevent participants from figuring out whether they were in the treatment or control group, e.g.: (a) the researchers didn’t tell the control group about the intervention, (b) placebos that looked the same as the real drug were used.

If the paragraph does not say anything about whether participants were blinded, say "n/a".

###

Paragraph: Between the 26th and 29th of April 1998, all households in the intervention villages received a number of Insecticide-treated bed nets (ITNs), according to the number of family members and were instructed explicitly about the correct use of the nets (Figure 2). Households in the control villages received identical-looking bed nets without insecticide treatment. More than 5000 ITN were distributed.

Quotes from paragraph that say whether the participants in the "Insecticide-treated bed nets (ITN)" intervention were blinded to allocation:
"Households in the control villages received identical-looking bed nets without insecticide treatment."

###

Paragraph: Between the 26th and 29th of April 1998, all households in the intervention villages received a number of Insecticide-treated bed nets (ITNs), according to the number of family members and were instructed explicitly about the correct use of the nets (Figure 2). Households in the control villages received identical-looking bed nets without insecticide treatment. More than 5000 ITN were distributed.

Quotes from paragraph that say whether the participants in the "meclizine" intervention were blinded to allocation:
n/a

###

Paragraph: The study was an open randomized controlled clinical trial. After a census of the village population, subjects in the target age group were screened for inclusion and exclusion criteria. Subjects who met inclusion and exclusion criteria were randomized either to receive two intermittent preventive treatments with standard recommended treatment doses of SP or no intermittent preventive treatment.

Quotes from paragraph that say whether the participants in the "intermittent preventive malaria treatment (IPT)" intervention with sulphadoxine pyrimethamine (SP) were blinded to allocation:
"The study was an open randomized controlled clinical trial."
"Subjects who met inclusion and exclusion criteria were randomized either to receive two intermittent preventive treatments with standard recommended treatment doses of SP or no intermittent preventive treatment."

###

Paragraph: To facilitate community-based treatment of malaria with the assigned regimen (artemether-lumefantrine or dihydroartemisinin-piperaquine), and to ensure that children received the correct regimen if they attended at health centres in the study area, ID cards were colour-coded according to intervention group and labelled with the regimen to be used for case management. The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded).

Quotes from paragraph that say whether the participants in the "short-acting ACT for case management of malaria (artemether-lumefantrine, AL) plus placebo SMC" intervention were blinded to allocation:
"The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded)."

###

Paragraph: To facilitate community-based treatment of malaria with the assigned regimen (artemether-lumefantrine or dihydroartemisinin-piperaquine), and to ensure that children received the correct regimen if they attended at health centres in the study area, ID cards were colour-coded according to intervention group and labelled with the regimen to be used for case management. The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded).

Quotes from paragraph that say whether the participants in the "finasteride" intervention were blinded to allocation:
n/a

###

Paragraph: In India, the Integrated Child Development Service (ICDS) maintains a network of child-care centres, caring for children up to age 6 years and off ering the potential to deliver simple health interventions. In rural Uttar Pradesh in north India, our plans to use the ICDS infrastructure for a large cluster-randomised trial of the eff ects of 6-monthly deworming with albendazole on pre-school child mortality were revised into plans for a factorial trial that would also evaluate the eff ects on mortality of enhancing vitamin A coverage. This 5-year trial of Deworming and Enhanced Vitamin A supple mentation (DEVTA) in 1 million pre-school children was larger than all other vitamin A trials combined. Its primary aim was to assess eff ects of a standard periodic treatment regimen on mortality at ages 1·0–6·0 years.

Quotes from paragraph that say whether the participants in the "6-monthly vitamin A" intervention were blinded to allocation:
n/a

###

"""

PERSONNEL_CONTEXT = """Instructions for each paragraph:

What does the paper say explicitly about whether the personnel (i.e. the people who carried out the intervention) were blinded to whether the participants were in the treatment or control group?

Include all relevant quotes:
- Everything the study explicitly says about whether personnel were blinded. If the paper says that the experiment was "open label", that means personnel knew whether participants were in the treatment or control group. So, the experiment was not blinded. We want to know about blinding of all personnel, including personnel who administered the intervention and personnel who collected data.
- All efforts that the research team made to prevent personnel from figuring out whether participants were in the treatment or control group. E.g. placebos that looked the same as the real drug were used, and personnel didn’t know which was the placebo.

If the paragraph does not say anything about whether participants were blinded, say "n/a".

###

Paragraph: Between the 26th and 29th of April 1998, all households in the intervention villages received a number of Insecticide-treated bed nets (ITNs), according to the number of family members and were instructed explicitly about the correct use of the nets (Figure 2). Households in the control villages received identical-looking bed nets without insecticide treatment. More than 5000 ITN were distributed.

Quotes from paragraph that say whether the personnel in the "Insecticide-treated bed nets (ITN)" intervention were blinded to allocation:
n/a

###

Paragraph: The study was an open randomized controlled clinical trial. Some study personnel knew which participants were in the treatment group. After a census of the village population, subjects in the target age group were screened for inclusion and exclusion criteria. Subjects who met inclusion and exclusion criteria were randomized either to receive two intermittent preventive treatments with standard recommended treatment doses of SP or no intermittent preventive treatment.

Quotes from paragraph that say whether the personnel in the "intermittent preventive malaria treatment (IPT)" intervention with sulphadoxine pyrimethamine (SP) were blinded to allocation:
"The study was an open randomized controlled clinical trial."
"Some study personnel knew which participants were in the treatment group."

###

Paragraph: In India, the Integrated Child Development Service (ICDS) maintains a network of child-care centres, caring for children up to age 6 years and off ering the potential to deliver simple health interventions. In rural Uttar Pradesh in north India, our plans to use the ICDS infrastructure for a large cluster-randomised trial of the eff ects of 6-monthly deworming with albendazole on pre-school child mortality were revised into plans for a factorial trial that would also evaluate the eff ects on mortality of enhancing vitamin A coverage. This 5-year trial of Deworming and Enhanced Vitamin A supple mentation (DEVTA) in 1 million pre-school children was larger than all other vitamin A trials combined. Its primary aim was to assess eff ects of a standard periodic treatment regimen on mortality at ages 1·0–6·0 years.

Quotes from paragraph that say whether the personnel in the "6-monthly vitamin A" intervention were blinded to allocation:
n/a

###

Paragraph: To facilitate community-based treatment of malaria with the assigned regimen (artemether-lumefantrine or dihydroartemisinin-piperaquine), and to ensure that children received the correct regimen if they attended at health centres in the study area, ID cards were colour-coded according to intervention group and labelled with the regimen to be used for case management. The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded).

Quotes from paragraph that say whether the personnel in the "short-acting ACT for case management of malaria (artemether-lumefantrine, AL) plus placebo SMC" intervention were blinded to allocation:
"The study was therefore open-label with respect to the regimen used for case management but blinded with respect to whether seasonal malaria chemoprevention was active or placebo (members of the research team from KNUST/CGHR and LSHTM were aware of the allocation, but those who administered the SMC, and mothers/children were blinded)."

###

Paragraph: "This was an open-label, randomised controlled trial enrolling asymptomatic schoolchildren and investigating the efficacy of IPTsc with sulfadoxine/pyrimethamine (SP) or SP combined with piperaquine (PQ) on anaemia and malaria morbidity in Congolese schoolchildren. The trial was composed of three treatment arms: SP monotherapy; SP combined with PQ (SP/PQ); and no antima- larial treatment (control). The study was conducted in Mokali health area, a semi-rural area of Kinshasa, Democratic Republic of the Congo (DRC). Malaria transmission is intense and perennial in this region, with two seasonal peaks (March–May and November–December) [19,20]."

Quotes from paragraph that say whether the participants in the "sulfadoxine/pyrimethamine" intervention were blinded to allocation:
"This was an open-label, randomised controlled trial enrolling asymptomatic schoolchildren and investigating the efficacy of IPTsc with sulfadoxine/pyrimethamine (SP) or SP combined with piperaquine (PQ) on anaemia and malaria morbidity in Congolese schoolchildren."

###

"""


def parse_quotes(s: str, ignore="n/a"):
    return [
        clean_line
        for line in s.split("\n")
        if (clean_line := line.strip('"\n ')) and clean_line != ignore
    ]


def make_paragraph_prompt(paragraph: Paragraph, intervention: str, group: Group) -> str:
    context = PARTICIPANTS_CONTEXT if group == "participants" else PERSONNEL_CONTEXT
    question = f"""Paragraph: {str(paragraph)}

Quotes from paragraph that say whether the {group} in the "{intervention}" intervention were blinded to allocation:"""
    return context + question


def make_followup_paragraph_prompt(
    paragraph: Paragraph, intervention: str, group: Group, answer: str
) -> str:
    prev_prompt = make_paragraph_prompt(paragraph, intervention, group)
    new_context = f"{prev_prompt} {answer}\n\n"
    new_question = f"""Other quotes from this paragraph relevant to whether the {group} in the "{intervention}" intervention were blinded to allocation:"""
    return new_context + new_question


class BlindingDynamic(Recipe):
    async def interventions(self, paper: Paper) -> list[str]:
        return list_experiments(document_id=paper.document_id)

    async def blinding_by_paragraph(
        self,
        paragraph: Paragraph,
        intervention: str,
        group: Group,
        gold_standard: GoldStandard,
        ask_followup: bool = False,
    ) -> ParagraphResult:
        """
        Does this paragraph state that [group] is blinded wrt
        [intervention]? Return an answer supported by a quote.
        """
        prompt = make_paragraph_prompt(paragraph, intervention, group)
        raw_quotes = await self.agent().complete(prompt=prompt)
        quotes = parse_quotes(raw_quotes, ignore="n/a")
        if quotes and ask_followup:
            followup_prompt = make_followup_paragraph_prompt(
                paragraph, intervention, group, raw_quotes
            )
            followup_quotes = await self.agent().complete(
                prompt=followup_prompt,
            )
            lines = parse_quotes(followup_quotes, ignore="n/a")
            quotes += lines
        for gold_quote in gold_standard.quotes:
            if gold_quote in str(paragraph):
                quotes_string = " ".join(quotes)
                if gold_quote not in quotes_string:
                    rich.print(
                        Panel(f"{prompt} {quotes_string}\n\nGold quote: {gold_quote}")
                    )
        # rich.print(Panel(f"{question} {quotes}"))
        return ParagraphResult(quotes=quotes, paragraph=paragraph)

    async def combine_paragraph_results(
        self,
        paper: Paper,
        paragraph_results: list[ParagraphResult],
        intervention: str,
        group: Group,
    ) -> InterventionResult:
        """
        Summarize whether [group] was blinded wrt [intervention] based on
        paragraph-level results.
        """
        return InterventionResult(
            document_id=paper.document_id,
            question_short_name=f"blinding_{group}_explicit",
            experiment=intervention,
            answer=None,
            quotes=list(set(flatten(result.quotes for result in paragraph_results))),
        )

    async def blinding_for_intervention(
        self, paper: Paper, intervention: str
    ) -> dict[Group, dict[str, Any]]:
        """
        For both participants and personnel, summarize whether they were
        blinded wrt [intervention].
        """
        results_by_group = {}
        groups: tuple[Group, Group] = ("participants", "personnel")
        for group in groups:
            gold_standard = get_blinding_gold_standards(
                paper=paper, group=group, intervention=intervention
            )
            paragraph_results = []
            for paragraph in tqdm.tqdm(paper.paragraphs):
                result = await self.blinding_by_paragraph(
                    paragraph, intervention, group, gold_standard=gold_standard
                )
                paragraph_results.append(result)
            group_results = await self.combine_paragraph_results(
                paper, paragraph_results, intervention, group
            )
            comparison = {
                "actual": group_results,
                "gold": gold_standard,
            }
            # Quick recall stats
            # pprint(comparison)
            recalled_quotes, missed_quotes = sort_by_recall(
                gold_standard.quotes, group_results.quotes
            )
            print(
                f"Recall for {paper.document_id} - {intervention} - {group}: {len(recalled_quotes)} / {len(gold_standard.quotes)} (with {len(group_results.quotes)} results)"
            )
            if missed_quotes:
                print(f"Missed quotes: {', '.join(quote(q) for q in missed_quotes)}")
            results_by_group[group] = comparison
        return results_by_group

    async def run(self, paper: Paper):
        """
        For each intervention, summarize how it was blinded.
        """
        results_by_intervention: dict[str, dict[Group, dict[str, Any]]] = {}
        interventions = await self.interventions(paper)
        for intervention in interventions:
            results_by_intervention[
                intervention
            ] = await self.blinding_for_intervention(paper, intervention)

        recipe_results: list[RecipeResult] = []
        for intervention in interventions:
            for group in results_by_intervention[intervention]:
                if group == "participants":
                    continue
                result: InterventionResult = results_by_intervention[intervention][
                    group
                ]["actual"]
                recipe_results.append(
                    RecipeResult(
                        document_id=result.document_id,
                        question_short_name=result.question_short_name,
                        experiment=result.experiment,
                        result=result,
                        excerpts=result.quotes,
                    )
                )
        self.maybe_add_to_results(recipe_results)
        return recipe_results



from ice.recipes.single_prompt import SinglePrompt

QUESTION_SHORT_NAME = "adherence"

DEFAULT_ANSWER_CLASSIFICATION: None = None

AI_PROMPT: str = "\nAssistant:"

HUMAN_PROMPT: str = "\nHuman:"

qa_prompt_template = f"""
{HUMAN_PROMPT} I'm trying to evaluate some RCTs. I've been told I should consider the adherence, attrition, and compliance of these papers when thinking about how much I should trust their results. Can you define these terms for me?
{AI_PROMPT} Sure, adherence describes how many participants selected for an intervention actually received it. Attrition describes how many of the initial sample dropped out of the study or were otherwise not available to be included in the final analysis. Compliance describes how well participants in the intervention complied with its protocol. All of these are important for the internal validity of a study, which informs how much we should trust its results.
{HUMAN_PROMPT} Here's the text of a paper I've been thinking about. Can you read it and summarize what it says, if anything, about adherence, attrition, and compliance?

{{paper_text}}
{AI_PROMPT} First, I'll identify all the parts of the paper that talk about adherence, attrition, or compliance. Then, I'll summarize these sections."""


class AdherenceSimpleInstruct(SinglePrompt):
    agent_str = "instruct"
    max_tokens = 3500
    qa_prompt_template = qa_prompt_template
    question_short_name = QUESTION_SHORT_NAME
    default_answer_classification = DEFAULT_ANSWER_CLASSIFICATION



from ice.recipe import recipe
from ice.recipes.primer.verify.last import check_step
from ice.recipes.primer.verify.utils import *
from ice.utils import map_async


async def verify_answer(
    question: str = DEFAULT_QUESTION, steps: list[str] = DEFAULT_STEPS
):
    """
    For each prefix of 1..n steps, check if the nth step is correct.
    """
    step_indices = list(range(1, len(steps) + 1))
    step_probs = await map_async(
        step_indices,
        lambda index: check_step(question=question, steps=steps[:index]),
    )
    return list(zip(step_probs, steps))


recipe.main(verify_answer)



from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Sequence
from typing import Optional
from typing import TypeVar

from structlog.stdlib import get_logger

from ice.apis.openai import openai_complete
from ice.apis.openai import TooLongRequestError
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt
from ice.recipes.experiments_and_arms.types import PassageWithReasoning
from ice.utils import map_async

log = get_logger()

T1 = TypeVar("T1")
T2 = TypeVar("T2")


async def reasoning_sample(
    prompt: str,
    get_reasoning: Callable[[str], str],
    temperature: float,
    stop: tuple[str, ...],
    cache_id: int,
) -> str:
    """Sample reasoning from the agent.

    Args:
        prompt (str): The reasoning prompt
        get_reasoning (Callable[[str], str]): Extract the reasoning from the generated answer.
        temperature (float): The temperature at which to sample.
        stop (tuple[str, ...]): Stop sequence(s) for reasoning
        cache_id (int): Cache ID for persistent cache identity across multiple calls.

    Returns:
        str: Generated reasoning.
    """
    response = await openai_complete(
        prompt=prompt,
        temperature=temperature,
        cache_id=cache_id,
        max_tokens=250,  # TODO: Reduce based on actual completion lengths
        stop=stop,
    )
    reasoning: str = response["choices"][0]["text"]
    return get_reasoning(reasoning)


async def greedy_continuation(
    texts: Sequence[str],
    helpfulness: Optional[str],
    reasoning: Optional[str],
    prompt_func: MultipartReasoningPrompt,
    max_tokens: int,
    final_answer_processor: Callable[[dict], T1],
) -> tuple[dict, T1]:
    """Greedily sample a continuation.

    Args:
        texts (Sequence[str]): The texts for the prompt.
        helpfulness (str | None): Generated helpfulness.
        reasoning (str | None): Generated reasoning.
        prompt_func (MultipartReasoningPrompt): Create a prompt from the texts, helpfulness (if it exists), and reasoning.
        max_tokens (int): Max tokens for the openai request.
        final_answer_processor (Callable[[dict], T1]): Function to extract an answer from the OpenAI response.

    Returns:
        tuple[dict, T1]: OpenAI response and the extracted answer.
    """
    prompt = prompt_func(texts, helpfulness=helpfulness, reasoning=reasoning)
    response = await openai_complete(
        prompt=prompt, logprobs=10, max_tokens=max_tokens, stop=None
    )
    return response, final_answer_processor(response)


async def _get_helpfulness(
    texts: Sequence[str],
    reasoning: str,
    prompt_func: MultipartReasoningPrompt,
    get_helpfulness: Callable[[str], str],
    max_tokens: int = 100,
) -> str:
    """Get helpfulness based on the generated reasoning.

    Args:
        texts (Sequence[str]): The texts for the prompt.
        reasoning (str): Generated reasoning.
        prompt_func (MultipartReasoningPrompt): Create a prompt from texts and reasoning.
        get_helpfulness (Callable[[str], str]): Extract helpfulness statement from generated answer.
        max_tokens (int, optional): Max tokens for the helpfulness request. Defaults to 100.

    Returns:
        str: The generated helpfulness.
    """
    _, helpfulness = await greedy_continuation(
        texts,
        helpfulness=None,
        reasoning=reasoning,
        prompt_func=prompt_func,
        max_tokens=max_tokens,
        final_answer_processor=lambda resp: get_helpfulness(resp["choices"][0]["text"]),
    )
    return helpfulness


async def _get_final_answer(
    texts: Sequence[str],
    helpfulness: Optional[str],
    reasoning: str,
    prompt_func: MultipartReasoningPrompt,
    final_answer_processor: Callable[[dict], T1],
    max_tokens: int = 50,
) -> T1:
    """Get final answer based on the generated reasoning (and optionally, helpfulness).

    Args:
        texts (Sequence[str]): The texts for the prompt.
        helpfulness (str | None): Generated helpfulness, if it exsts.
        reasoning (str): Generated reasoning.
        prompt_func (MultipartReasoningPrompt): Create a prompt from the texts, helpfulness (if it exists), and reasoning.
        final_answer_processor (Callable[[dict], T1]): Extract final answer from the OpenAI response.
        max_tokens (int, optional): Max tokens for the final answer generation request. Defaults to 50.

    Returns:
        T1: The final answer.
    """
    _, final_answer = await greedy_continuation(
        texts,
        helpfulness=helpfulness,
        reasoning=reasoning,
        prompt_func=prompt_func,
        max_tokens=max_tokens,
        final_answer_processor=final_answer_processor,
    )
    return final_answer


async def reason_select_and_answer(
    texts: Sequence[str],
    num_examples: int,
    cache_id: int,
    reasoning_temperature: float,
    reasoning_stop: tuple[str, ...],
    prompt_func: Callable[[int], MultipartReasoningPrompt],
    get_reasoning: Callable[[str], str],
    get_helpfulness: Optional[Callable[[str], str]],
    final_answer_processor: Callable[[dict], T1],
) -> PassageWithReasoning[T1]:
    """Prompt chaining technique that samples reasoning, then optionally greedily generates helpfulness, then a final answer.

    Args:
        texts (Sequence[str]): The texts for the prompt.
        num_examples (int): The number of shots for few-shot prompts.
        cache_id (int): Cache ID for cache-breaking purposes.
        reasoning_temperature (float): Temperature for reasoning sample.
        reasoning_stop (tuple[str, ...]): Stop sequence(s) for reasoning.
        prompt_func (Callable[[int], MultipartReasoningPrompt]): Generate prompts based on texts, (optional) reasoning, (optional) helpfulness, and the final answer.
        get_reasoning (Callable[[str], str]): Extract reasoning from the generation in response to the reasoning prompt.
        get_helpfulness (Callable[[str], str] | None): Extract helpfulness from the generation in response to the helpfulness prompt. If None, do not generate helpfulness.
        final_answer_processor (Callable[[dict], T1]): Extract final answer from the OpenAI response to final answer prompt.

    Raises:
        TooLongRequestError: If the prompts are too long after first reducing the number of shots, then dropping texts.

    Returns:
        PassageWithReasoning[T1]: Final answer with reasoning.
    """
    try:
        reasoning_prompt = prompt_func(num_examples)([t for t in texts])
        reasoning = await reasoning_sample(
            reasoning_prompt,
            get_reasoning,
            reasoning_temperature,
            reasoning_stop,
            cache_id,
        )
        if get_helpfulness:
            helpfulness = await _get_helpfulness(
                texts,
                reasoning=reasoning,
                prompt_func=prompt_func(num_examples),
                get_helpfulness=get_helpfulness,
            )
        else:
            helpfulness = None
        final_answer = await _get_final_answer(
            texts,
            helpfulness=helpfulness,
            reasoning=reasoning,
            prompt_func=prompt_func(num_examples),
            final_answer_processor=final_answer_processor,
        )
        return PassageWithReasoning(
            passage=texts,
            reasoning=reasoning,
            helpfulness=helpfulness,
            final_answer=final_answer,
        )

    except TooLongRequestError as e:
        try:
            if num_examples > 1:
                num_examples -= 1
                log.warn(
                    "Trying shorter prompt",
                    num_examples=num_examples,
                    error_detail=e.detail,
                )
                return await reason_select_and_answer(
                    texts=texts,
                    num_examples=num_examples,
                    cache_id=cache_id,
                    reasoning_temperature=reasoning_temperature,
                    reasoning_stop=reasoning_stop,
                    prompt_func=prompt_func,
                    get_reasoning=get_reasoning,
                    get_helpfulness=get_helpfulness,
                    final_answer_processor=final_answer_processor,
                )
            else:
                raise e
        except TooLongRequestError as e:
            if texts:
                texts = texts[:-1]
                log.warn("Dropping texts", num_texts=len(texts), error_detail=e.detail)
                return await reason_select_and_answer(
                    texts=texts,
                    num_examples=num_examples,
                    cache_id=cache_id,
                    reasoning_temperature=reasoning_temperature,
                    reasoning_stop=reasoning_stop,
                    prompt_func=prompt_func,
                    get_reasoning=get_reasoning,
                    get_helpfulness=get_helpfulness,
                    final_answer_processor=final_answer_processor,
                )
            else:
                raise e


async def answer_with_best_reasoning(
    num_samples: int,
    selector: Callable[[Sequence[PassageWithReasoning[T1]]], Awaitable[T2]],
    texts: Sequence[str],
    num_examples: int,
    reasoning_temperature: float,
    reasoning_stop: tuple[str, ...],
    prompt_func: Callable[[int], MultipartReasoningPrompt],
    get_reasoning: Callable[[str], str],
    get_helpfulness: Optional[Callable[[str], str]],
    final_answer_processor: Callable[[dict], T1],
) -> T2:
    """Sample multiple reasonings, greedily generating helpfulness (optional) and a final answer based on the reasoning.
    Choose a final answer based on the technique passed in as the `selector`.

    Args:
        num_samples (int): Number of distinct reasoning candidates to generate.
        selector (Callable[[Sequence[PassageWithReasoning[T1]]], Awaitable[T2]]): Function that selects the "best" reasoning among the candidates.
        texts (Sequence[str]): Texts for the prompts.
        num_examples (int): Number of shots for the prompts (will be reduced if prompts are too long).
        reasoning_temperature (float): Reasoning temperature.
        reasoning_stop (tuple[str, ...]): Stop sequence(s) for the reasoning
        prompt_func (Callable[[int], MultipartReasoningPrompt]): Given the number of shots, return a function that can generate prompts based on the chaining technique applied here.
        get_reasoning (Callable[[str], str]): Extract reasoning from the reasoning prompt.
        get_helpfulness (Callable[[str], str] | None): Extract helpfulness from the helpfulness prompt. If `None`, do not generate helpfulness.
        final_answer_processor (Callable[[dict], T1]): Extract final answer (generic) based on the OpenAI response to the greedy final answer prompt.

    Returns:
        T2: Final answer.
    """

    async def sample(cache_id: int):
        return await reason_select_and_answer(
            texts,
            num_examples=num_examples,
            cache_id=cache_id,
            reasoning_temperature=reasoning_temperature,
            reasoning_stop=reasoning_stop,
            prompt_func=prompt_func,
            get_reasoning=get_reasoning,
            get_helpfulness=get_helpfulness,
            final_answer_processor=final_answer_processor,
        )

    answers = await map_async(
        [cache_id for cache_id in range(num_samples)],
        sample,
        max_concurrency=10,
        show_progress_bar=False,
    )

    return await selector(answers)


recipe.main(answer_with_best_reasoning)



from collections.abc import Sequence
from dataclasses import dataclass
from statistics import mean
from typing import Optional
from typing import Union

import pandas as pd

from pydantic import BaseModel
from rich import box
from rich.table import Table

from ice.evaluation.evaluate_recipe_result import EvaluatedRecipeResult
from ice.evaluation.utils import CSVS_PATH
from ice.evaluation.utils import precision_score
from ice.evaluation.utils import recall_score
from ice.evaluation.utils import start_time
from ice.utils import latest_commit_hash

SUBSTANTIVELY_CORRECT_MIN_ANSWER_RATING = 4


def as_percent(decimal: Optional[float]) -> Optional[str]:
    if decimal is None:
        return None
    return "{:.0%}".format(decimal)


@dataclass
class ClassificationSummary:
    num_correct: int
    num_evaluated: int
    classification_options: set[str]

    @property
    def proportion_correct(self) -> Optional[float]:
        if self.num_evaluated == 0:
            return None

        return self.num_correct / self.num_evaluated

    def labels_str(self) -> str:
        return ", ".join(list(self.classification_options))


@dataclass
class AnsweredWhenShouldHaveSummary:
    gs_should_answer: list[bool]
    answered: list[bool]

    @property
    def precision(self) -> float:
        if not self.gs_should_answer:
            return 0.0
        return precision_score(self.gs_should_answer, self.answered)

    @property
    def recall(self) -> float:
        if not self.gs_should_answer:
            return 0.0
        return recall_score(self.gs_should_answer, self.answered)

    @property
    def proportion_of_gs_where_should_answer(self) -> float:
        return len(
            [should_answer for should_answer in self.gs_should_answer if should_answer]
        ) / (len(self.gs_should_answer) or 1)


class EvaluationReport(BaseModel):
    technique_name: str
    results: Sequence[EvaluatedRecipeResult]

    @property
    def mean_proportion_gs_found_across_papers(self) -> Optional[float]:
        paper_scores = [
            result.evaluated_excerpts.proportion_gold_standards_found
            for result in self.results
            if result.evaluated_excerpts.proportion_gold_standards_found is not None
        ]

        if len(paper_scores) == 0:
            return None
        return float(mean(paper_scores))

    @property
    def proportion_papers_with_gs_found(self) -> Optional[float]:
        gs_found = [
            result.evaluated_excerpts.num_gold_standards_found > 0
            for result in self.results
            if result.evaluated_excerpts.num_gold_standards_found is not None
        ]

        if len(gs_found) == 0:
            return None

        return mean(gs_found)

    @property
    def classification_summaries(self) -> list[ClassificationSummary]:
        classification_summaries = []

        for i in range(
            0,
            max([len(result.evaluated_classifications) for result in self.results]),
        ):
            is_corrects = []

            for result in self.results:
                try:
                    is_corrects.append(result.evaluated_classifications[i].is_correct)
                except IndexError:
                    continue

            evaluations = [
                is_correct for is_correct in is_corrects if is_correct is not None
            ]

            classification_summaries.append(
                ClassificationSummary(
                    num_correct=sum(evaluations),
                    num_evaluated=len(evaluations),
                    classification_options=self.classification_options(i),
                )
            )

        return classification_summaries

    @property
    def answered_when_should_have(self) -> AnsweredWhenShouldHaveSummary:
        results_to_eval = [
            result for result in self.results if result.gold_standard is not None
        ]
        gs_should_answer = [
            False if result.gold_standard.answer == "" else True  # type: ignore
            for result in results_to_eval
        ]
        answered = [
            False if result.answer == "" else True for result in results_to_eval
        ]

        return AnsweredWhenShouldHaveSummary(
            gs_should_answer=gs_should_answer,
            answered=answered,
        )

    @property
    def elicit_commit(self):
        elicit_commits = list(set([result.elicit_commit for result in self.results]))
        if len(elicit_commits) > 1:
            raise ValueError("Attempting to evaluate multiple Elicit commits at once")
        return elicit_commits[0]

    @property
    def answer_ratings(self):
        return [
            result.answer_rating
            for result in self.results
            if result.answer_rating is not None
        ]

    @property
    def answers_rated(self):
        return len(self.answer_ratings)

    @property
    def average_answer_rating(self):
        return mean(self.answer_ratings) if len(self.answer_ratings) > 0 else None

    @property
    def proportion_of_answers_rated_correct(self):
        return (
            (
                len(
                    [
                        answer_rating
                        for answer_rating in self.answer_ratings
                        if answer_rating >= SUBSTANTIVELY_CORRECT_MIN_ANSWER_RATING
                    ]
                )
                / len(self.answer_ratings)
            )
            if self.answer_ratings
            else None
        )

    @property
    def failure_modes(self):
        # TODO: use more_itertools.flatten instead
        all_failure_modes = []
        for result in self.results:
            if result.failure_modes is None:
                continue
            for failure_mode in result.failure_modes:
                all_failure_modes.append(failure_mode)

        if not all_failure_modes:
            return None

        return pd.Series(all_failure_modes).value_counts().to_string()

    def classification_options(self, classification_idx=int) -> set[str]:
        classification_options: set[str] = set()

        for result in self.results:
            try:
                gold_classification = result.evaluated_classifications[
                    classification_idx
                ].gold
                if gold_classification is not None:
                    classification_options.add(gold_classification)
            except IndexError:
                continue

        return classification_options

    def questions_str(self) -> str:
        questions = list(set([result.question_short_name for result in self.results]))
        return ", ".join(questions)

    def classification_summary_stats_str(self) -> str:
        return "\n".join(
            [
                (
                    f"- Classification {i+1}:"
                    + "\n    - proportion correct: "
                    + (
                        f"{classification_summary.proportion_correct:.0%} ({classification_summary.num_correct} / {classification_summary.num_evaluated})"
                        if classification_summary.proportion_correct is not None
                        else "Nothing to evaluate"
                    )
                    + f"\n    - labels: {classification_summary.labels_str()}"
                )
                for i, classification_summary in enumerate(
                    self.classification_summaries
                )
            ]
        )

    def make_comparison_table(
        self,
        document_id_rows: list[str],
        gold_standard_rows: list[str],
        model_rows: list[str],
    ) -> Table:
        table = Table(width=120, box=box.MARKDOWN)
        table.add_column("Document", no_wrap=True)
        table.add_column("Gold Standard")
        table.add_column("Model")

        for did, gs, mr in zip(document_id_rows, gold_standard_rows, model_rows):
            table.add_row(str(did), str(gs), str(mr))

        return table

    def description_results_table(self) -> Table:
        document_ids = []
        gold_standard_descriptions = []
        model_descriptions = []
        for result in self.results:
            gold_standard = result.gold_standard
            if gold_standard is not None:
                if gold_standard.answer != gold_standard.classifications[0]:
                    document_ids.append(gold_standard.document_id)
                    gold_standard_descriptions.append(gold_standard.answer)
                    model_descriptions.append(result.answer)
        return self.make_comparison_table(
            document_ids, gold_standard_descriptions, model_descriptions
        )

    def classification_results_table(self) -> Table:
        document_ids = []
        gold_standard_classifications = []
        model_classifications = []
        for result in self.results:
            gold_standard = result.gold_standard
            if gold_standard is not None:
                document_ids.append(gold_standard.document_id)
                gold_standard_classifications.append(
                    gold_standard.classifications[1] or ""
                )
                model_classification = (
                    result.classifications[0] if result.classifications else ""
                )
                model_classifications.append(model_classification or "")
        return self.make_comparison_table(
            document_ids, gold_standard_classifications, model_classifications
        )

    def excerpts_summary_stats_str(self) -> str:
        s = ""
        if self.mean_proportion_gs_found_across_papers:
            s += f"### Average % of gold standards found: \n\n{self.mean_proportion_gs_found_across_papers:.0%}"
        if self.proportion_papers_with_gs_found:
            s += f"\n\n### Proportion of papers with at least one gold standard found: \n\n{self.proportion_papers_with_gs_found:.0%}"
        if not s:
            s = "n/a"
        return s

    def precision_recall_str(self) -> str:
        return f"""- precision: {self.answered_when_should_have.precision:.0%}
- recall: {self.answered_when_should_have.recall:.0%}
- (proportion of gold standards where should answer: {self.answered_when_should_have.proportion_of_gs_where_should_answer:.0%})
"""

    def to_rich_elements(self) -> list[Union[str, Table]]:
        return [
            f"""
# Performance on papers

{"---".join(str(r) for r in self.results)}

# Summary of results

## Classifications:
""",
            self.classification_results_table(),
            """
## Descriptions:
""",
            self.description_results_table(),
            """
## Classification summary stats:
""",
            self.classification_summary_stats_str(),
            """
## Excerpts summary stats:
""",
            self.excerpts_summary_stats_str(),
            """
## Answered when should have:
""",
            self.precision_recall_str(),
        ]

    def __str__(self) -> str:
        return "\n".join(str(e) for e in self.to_rich_elements())

    def make_dashboard_row_df(self):
        CSVS_PATH.mkdir(parents=True, exist_ok=True)

        questions = self.questions_str()
        row = {
            "Questions": questions,
            "Technique": self.technique_name,
            "Eval date": start_time,
            "Elicit commit": self.elicit_commit,
            "ICE commit": latest_commit_hash(),
            "Splits": ", ".join(
                set(
                    result.gold_standard.split
                    for result in self.results
                    if result.gold_standard and result.gold_standard.split
                )
            ),
            "# of results": len(self.results),
            "Papers with >= 1 GS quote found": as_percent(
                self.proportion_papers_with_gs_found
            ),
            "Mean % of GS quotes found": as_percent(
                self.mean_proportion_gs_found_across_papers
            ),
            "Precision (answer if should)": as_percent(
                self.answered_when_should_have.precision
            ),
            "Recall (answer if should)": as_percent(
                self.answered_when_should_have.recall
            ),
            "% GS where should answer": as_percent(
                self.answered_when_should_have.proportion_of_gs_where_should_answer
            ),
            "Answers rated": self.answers_rated,
            "Average answer rating": self.average_answer_rating,
            "% answers rated substantively correct": as_percent(
                self.proportion_of_answers_rated_correct
            ),
            "Failure mode counts": self.failure_modes,
        }
        for i, classification_summary in enumerate(self.classification_summaries):
            row[f"Classification {i+1} labels"] = classification_summary.labels_str()

            row[f"Classification {i+1} correct"] = as_percent(
                classification_summary.proportion_correct
            )

            row[
                f"Classification {i+1} # evaluated"
            ] = classification_summary.num_evaluated

        df = pd.DataFrame([row])
        df.to_csv(
            CSVS_PATH / f"dashboard_row {start_time} {questions}.csv",
            index=False,
        )
        return df

    def make_experiments_evaluation_df(self):
        CSVS_PATH.mkdir(parents=True, exist_ok=True)

        rows = []
        questions = self.questions_str()

        for result in self.results:
            row = {
                "question_short_name": result.question_short_name,
                "technique": self.technique_name,
                "eval_date": start_time,
                "elicit_commit": result.elicit_commit,
                "ice_commit": latest_commit_hash(),
                "document_id": result.document_id,
                "split": result.gold_standard.split if result.gold_standard else None,
                "experiment": result.gold_standard.experiment
                if result.gold_standard
                else None,
                "total_gs_quotes": len(
                    result.evaluated_excerpts.gold_standards_in_excerpts_results
                ),
                "gs_quotes_found": result.evaluated_excerpts.num_gold_standards_found,
                "proportion_gs_quotes_found": result.evaluated_excerpts.proportion_gold_standards_found,
                "evaluated_excerpts": str(result.evaluated_excerpts),
                "excerpts": result.evaluated_excerpts.excerpts,
                "gs_quotes": result.evaluated_excerpts.gold_standards_str(),
                "answer": result.answer,
                "gs_answer": result.gold_standard.answer
                if result.gold_standard
                else None,
                "answer_rating": result.answer_rating,
                "failure_modes": result.failure_modes,
            }

            if result.gold_standard:
                row["gs_answer"] = result.gold_standard.answer
                row["should_answer"] = result.gold_standard.answer != ""
                row["answered"] = result.answer != ""

            for i, classification in enumerate(result.evaluated_classifications):
                row[f"classification_{i+1}_correct"] = classification.is_correct
                row[f"classification_{i+1}"] = classification.predicted
                row[f"gold_classification_{i+1}"] = classification.gold

            rows.append(row)
        df = pd.DataFrame(rows)
        df.to_csv(
            CSVS_PATH / f"experiments_evaluation {start_time} {questions}.csv",
            index=False,
        )
        return df



from collections.abc import Sequence

from ice.recipe import recipe
from ice.recipes.experiments_and_arms.num_utils import strip_enumeration_prefix
from ice.recipes.experiments_and_arms.prompts.cluster import build_cluster_prompt
from ice.recipes.experiments_and_arms.prompts.cluster import build_count_prompt
from ice.recipes.experiments_and_arms.prompts.cluster import build_final_prompt


async def best_answer_by_clustering(question: str, candidates: Sequence[str]) -> str:
    """Prompt chaining recipe that asks the agent to first identify clusters,
    then count the number of answers in each cluster, then summarize the
    best answer based on these clusters.

    Args:
        question (str): Question for the candidate answers.
        candidates (Sequence[str]): The candidate answers.

    Returns:
        str: The best answer identified by applying the clustering method.
    """
    clusters = await recipe.agent().complete(
        prompt=build_cluster_prompt(question, candidates)
    )
    counts = await recipe.agent().complete(
        prompt=build_count_prompt(question, candidates, clusters)
    )
    final_answer = await recipe.agent().complete(
        prompt=build_final_prompt(question, candidates, clusters, counts)
    )

    return strip_enumeration_prefix(final_answer)


recipe.main(best_answer_by_clustering)



from collections.abc import Callable
from collections.abc import Sequence
from typing import Optional
from typing import Union

from ice.formatter.multi import format_multi
from ice.formatter.transform.dependent import CountWord
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import PositionalTransform
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.prompts.utils import start_last_example
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt


class FirstOrNow(PositionalTransform):
    def transform(self, position: int, total: int) -> str:
        total  # ignored
        if position == 0:
            return "First,"
        else:
            return "Now"


class ExperimentOrExperiments(DependentTransform[list]):
    def key(self):
        return "paragraphs"

    def transform(self, dependent) -> str:
        return "experiment" if len(dependent) == 1 else "experiments"


CAN_WE_NAME_EXPERIMENTS_EXAMPLES: list[
    dict[str, Union[ValueTransform[Sequence[str]], str, int]]
] = [
    dict(
        paragraphs=numbered_list(
            [
                """We present results from six randomized control trials of an integrated approach to improve livelihoods amongst the very poor. The approach combines the transfer of a productive asset with consumption support, training and coaching plus savings encouragement and health education and/or services. Results from the implementation of the same basic program, adapted to a wide variety of geographic and institutional contexts and with multiple implementing partners, show statistically significant, costeffective impacts on consumption (fueled mostly by increases in self-employment income) and psychosocial status of the targeted households. The impact on the poor households lasted at least a year after all implementation ended. It is possible to make sustainable improvements in the economic status of the poor with a relatively short-term intervention.""",
                """More than one fifth of the world\'s population lives on less than Purchasing Power Parity (PPP) US$1.25 a day, and there is an emerging international consensus that this share should (and can) be driven close to zero by 2030 (1,2). Reaching this objective will require enabling the poorest families, who are often the most marginalized within their villages, to shift from insecure and fragile sources of income to more sustainable income-generating activities. One possible avenue, popular with both development organizations and governments, is to promote self-employment activities (such as cow rearing or petty trading). Past efforts to reduce poverty by encouraging these types of activities among the poor, however, have often been plagued by implementation problems and been deemed failures (3). For example, India\'s Integrated Rural Development Program is believed to have been both poorly targeted and ineffective (4,5). However, in recent years, several large non-governmental organizations (prominent international northern NGOs such as Oxfam, World Vision and Heifer, as well as many local NGOs) have gone back to this "livelihood" approach. This past experience raises the question: is it actually possible to reliably improve the livelihoods of the poorest households by giving them access to self-employment activities, or is this entire approach flawed? In particular, is it possible to come up with a model for doing so that can be implemented by a wide variety of organizations and works in a wide range of geographic, institutional, and cultural contexts?""",
                """We present results from randomized control trials (RCTs) in six countries of a particular approach to foster self-employment activities amongst the very poor. Originally designed and implemented by BRAC, a large Bangladeshi NGO that runs several country-wide programs, the "Graduation" program provides a holistic set of services, including the grant of a productive asset, to the poorest households in a village (referred to by BRAC as the "ultra-poor"). The beneficiaries are identified through a participatory process in a village meeting, followed by a verification visit by the organization\'s staff. Selected beneficiaries are then given a productive asset that they choose from a list, training and support for the asset they have chosen, as well as general life skills coaching, weekly consumption support for some fixed period, and typically access to savings accounts and health information or services. These different activities (plus regular interactions with the households over the course of a year) are designed to complement each other in helping households to start a productive self-employment activity. The idea is to provide a "big push", over a limited period of time, with the hope of unlocking a poverty trap. The program costs per household average 100% (range from 62% to 145%) of baseline household consumption. While the program may initially be relatively expensive (compared to just providing training, coaching or a cash transfer), the thinking behind the program is that the combination of these activities is necessary and sufficient to obtain persistent impact on a large fraction of the beneficiaries. 4 including one year after the end of the program, which directly speaks to the sustainability of the changes we observe.""",
                """Of the six experiments, three are individual randomized trials with randomization at the household level within each village (India, Ethiopia and Pakistan) and three are clustered randomized trials, with randomization at both the village and household level (Ghana, Honduras, and Peru). In the countries with clustered randomization, villages were randomly selected to be treatment or control villages, and then treatment households were randomly selected within the set of eligible households in treatment villages.""",
            ]
        ),
        reasoning="\n\n".join(
            [
                "Excerpt 1 does not name the six different RCTs.",
                "Excerpt 2 provides background information but does not name the experiments.",
                "Excerpt 3 indicates that the 6 RCTs were conducted in 6 different countries and descrbies what they have in common, which is to study the 'Graduation' program in each of these countries",
                "Excerpt 4 actually names the 6 experiments: 'India', 'Ethiopia', 'Pakistan', 'Ghana', 'Honduras', and 'Peru'.",
            ]
        ),
        helpfulness="Excerpt 3 was somewhat helpful, excerpt 4 was extremely helpful, and excerpts 1 and 2 were not helpful.",
        experiments_count=6,
        answer="Yes",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """We find that the effects of the workshop are significantly larger for the most disadvantaged. In particular, we show in Table 5 that the least educated, the least experienced, and those with the lowest expected earnings benefit the most from the interventions. For other dimensions, we are unable to find significant differences in response to treatment. 51  he size of the effects for the worst-off workers is substantial. For example, young people without tertiary education increase the earnings by almost 60 percent, while the low predicted earnings group experiences a 50 percent increase. This causes a large reduction in earning inequality: the earning gap between the low and the high earnings group drops from 142 percent to 54 percent and, strikingly, the gap between experienced and inexperienced workers is fully erased. Overall, these results illustrate the large equity gains that can be generated by helping young workers to access the labour market through improved signalling.""",
                """A.17""",
                """Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on financial outcomes. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1. Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on expectations, aspirations and reservation wages. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1. Note. In this table we report the intent-to-treat estimates of the direct and indirect effects of the transport intervention and the job application workshop on outcomes related to mobility. These are obtained by OLS estimation of equation ( 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1.""",
                """A.19 1), weighting each observation by the inverse of the probability of being sampled. Below each coefficient estimate, we report the s.e. in parentheses and the q-value in brackets. We correct standard errors to allow for arbitrary correlation at the level of geographical clusters. q-values are obtained using the sharpened procedure of Benjamini et al. (2006). Changing number of observations due to missing values in the dependent variable. In the last three columns we report the mean outcome for the control group, the p-value from a F-test of the null hypothesis that transport subsidies and the job application workshop have the same effect, and the number of observations. ***p< 0.01, **p<0.05, *p<0.1.""",
            ]
        ),
        reasoning="\n\n".join(
            (
                "Excerpt 1 discusses findings related to a workshop; this suggests that the experiment had to do with studying the effects of a workshop.",
                "Excerpt 2 is just a reference and does not have any useful content.",
                "Excerpt 3 is a note explaining some analytic methodology related to reporting outcomes from the job application workshop and the transport subsidy. Since there was a single experiment, this suggests that the experiment studied both a job application workshop and a transport subsidy.",
                "Excerpt 4 also explains some analytic methodology; because it seems to compare the effect sizes of the job application workshop and the transport subsidy together, it suggests that these might be two trial arms belonging to a single experiment, but it is ultimately unclear here.",
            )
        ),
        helpfulness="Excerpt 1 was somewhat helpful, excerpts 2 was not helpful, and excerpts 3 and 4 were moderately helpful.",
        experiments_count=1,
        answer="Yes",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """At the beginning of each study session, a verbal explanation about the study and the consent form were given in Nepali by the research assistant, and written informed consent was gathered from participants in a forward and back translated document presented in Nepali. During debriefing, participants were informed that they could still withdraw from the study by contacting the research staff without penalty. A mental health resource list was emailed to participants if they had further questions, concerns, or interest in mental health services. No participants self-disclosed experiencing distress during the procedure, so none were referred for mental healthcare during the study. Participants were not compensated for their participation in the research activities. The procedures for this study were approved by the ethical review boards at Duke University (E0078), TU-IOM (380), KUSMS, and the Nepal Health Research Council (146/2017) prior to data collection. All data were stored on an encrypted server and validated by a second investigator before performing analyses.""",
                """Descriptive statistics summarized the characteristics of the participants. The primary inferential analyses evaluated differences in attitudes and knowledge by comparing intervention groups. Differences in attitudes and knowledge between study conditions were analyzed using linear and logistic regression. The primary comparisons were between (a) both interventions compared to the control and (b) the two intervention groups to each other. Pearson correlation was used to identify relationships between implicit and explicit attitudes as well as correlations between attitudes and knowledge. The implicit and explicit correlation was done to test if reported and unconscious beliefs were related. The other correlation tested if greater knowledge was associated with more positive attitudes. Analyses using IAT data only included frequent computer users. A subsequent regression model was adjusted for personal experience with mental illness to determine the association between the intervention videos and the primary outcome. We conducted an exploratory analysis comparing findings from Study 1 (depression only) and Study 2 (depression and psychosis). All analyses were performed in STATA software v. 15.0 [51] with two-tailed tests using a significance level of 0.05.""",
                """For the qualitative component, interviews were audio recorded and transcribed for analysis. Data management and coding were facilitated using QSR NVivo 11 software [52]. Data analysis was guided using a content analysis strategy. First, the interviews were transcribed, and then two independent reviewers wrote memos for each of the transcripts to identify salient themes and potential codes. A codebook was created using inductive and deductive sources of information to generate major and minor categories of themes. Each code under a theme had a unique definition and inclusion/exclusion criteria. Both researchers applied the codes to each transcript independently, discussed discrepancies until consensus was reached, and iteratively modified the codebook after each discussion. Codes were merged and sub-categorized into different themes for optimal organization and fit. All the transcripts were coded through this technique.""",
                """After review of 2nd and 3rd year MBBS student rosters, 18 students were excluded prior to randomization because of being international students not speaking Nepali or having already completed their psychiatry rotation. Among the remaining students, 100 were selected for randomization to one of the three arms. No potential participants refused to participate in this study. An additional six students were excluded at the time of analysis because information on their demographic forms revealed that they were international students whose native language was not Nepali or they had completed their clinical psychiatry rotation; this information had not been up to date in the class rosters at the time of randomization (Figure 1 ). One participant in the service user arm was excluded because of both being an international non-Nepali student and having completed a psychiatry rotation. Demographic characteristics of these participants are in Table 2 . Participants were randomized into one the three conditions: the control group with no video (n = 31, 33%), the didactic video group (n = 31, 33%), and the service user recovery testimonial video group (n = 32; 34%).""",
            ]
        ),
        reasoning="\n\n".join(
            (
                """Excerpt 1 discusses study methodology, including informed consent and approval from ethical review boards, but it does not discuss how many experiments were conducted.""",
                """Excerpt 2 discusses the analyses methods applied, referring to two studies. Because they call these "studies" and not merely "groups," "conditions," or "arms", this indicates that this paper covers two distinct experiments, one "depression only" study and one "depression and psychosis" study.""",
                """Excerpt 3 discusses the qualitative component of the analysis but does not provide more information about the nature of the two experiments.""",
                """Excerpt 4 discusses randomization across three arms but does not provide more information about which experiment these arms belong to or what distingishes the two experiments.""",
            )
        ),
        helpfulness="Excerpts 1, 3, and 4 were not helpful. Excerpt 2 was somewhat helpful.",
        experiments_count=2,
        answer="Yes",
    ),
]

SHARED_CAN_WE_NAME_EXPERIMENTS_ARGS = dict(
    maybe_plural=plural_transform("experiments_count", "", "s"),
    first_or_now=FirstOrNow(),
    just_one_or_how_many=CountWord("experiments_count", {1: "just one"}),
    do_or_does_the_above_excerpts=plural_transform(
        "paragraphs", "does the above excerpt", "do the above excerpts"
    ),
    each_or_the=plural_transform("experiments_count", "the", "each"),
    were_or_was=plural_transform("experiments_count", "was", "were"),
    maybe_plural_excerpts=plural_transform("paragraphs", "", "s"),
)

CAN_WE_NAME_EXPERIMENTS_TEMPLATE = """{first_or_now} consider the following excerpt{maybe_plural_excerpts} from a research paper that conducted {just_one_or_how_many} experiment{maybe_plural}

{paragraphs}

Let's think it over:

{reasoning}

Which excerpts, if any, were helpful in understanding what the experiment{maybe_plural} conducted in this study {were_or_was}? {helpfulness}

Therefore, {do_or_does_the_above_excerpts} provide insight into what the experiment{maybe_plural} {were_or_was}, including the name of {each_or_the} experiment? If we cannot tell from the excerpt{maybe_plural_excerpts}, answer "Unclear". Answer Yes, No, or Unclear. {answer}
"""


def can_we_name_experiments_stop_seq(reasoning: Optional[str]) -> Sequence[str]:
    return ["\n\n"] if reasoning else ["Therefore, "]


def make_can_we_name_experiments_prompt(
    num_experiments: int,
) -> Callable[[int], MultipartReasoningPrompt]:
    def from_num_shot(num_shots: int):
        def can_we_name_experiments_prompt(
            paragraphs: Sequence[str],
            helpfulness: Optional[str] = None,
            reasoning: Optional[str] = None,
        ) -> str:
            last_example: dict = start_last_example(
                helpfulness=helpfulness, reasoning=reasoning, pre_final="Excerpt 1"
            ) | dict(
                paragraphs=numbered_list(paragraphs),
                experiments_count=num_experiments,
            )
            shots = format_multi(
                CAN_WE_NAME_EXPERIMENTS_TEMPLATE,
                CAN_WE_NAME_EXPERIMENTS_EXAMPLES[:num_shots] + [last_example],
                SHARED_CAN_WE_NAME_EXPERIMENTS_ARGS,
            )
            instructions = """The following excerpts are from a research papers that conducted different numbers of experiments. Experiments are separate studies on distinct populations. Each experiment may have multiple trial arms, which describe the different treatment or control conditions participants were in. Do the following excerpts describe what this experiment was? Don't confuse describing the experiment with describing trial arms in a single experiment."""
            parts = [instructions] + list(shots)
            return "\n\n".join(parts)

        return can_we_name_experiments_prompt

    return from_num_shot


CAN_WE_NAME_EXPERIMENTS_CHOICES = [" Yes", " No", " Unsure"]
CAN_WE_NAME_EXPERIMENTS_BEST_CHOICE = " Yes"
CAN_WE_NAME_EXPERIMENTS_REASONING_STOP = ("\nWhich excerpt",)


def get_can_we_name_experiments_reasoning(response: str) -> str:
    return "".join(
        ("Excerpt 1 ", get_part(response, "think it over:\n", "\nWhich excerpts"))
    )


def get_can_we_name_experiments_helpfulness(response: str) -> str:
    return get_part(response, "conducted in this study?", "Therefore, do")



from collections.abc import Mapping
from typing import Optional
from typing import Union

import httpx

from httpx import Response
from httpx import TimeoutException
from structlog.stdlib import get_logger
from tenacity import retry
from tenacity.retry import retry_any
from tenacity.retry import retry_if_exception
from tenacity.retry import retry_if_exception_type
from tenacity.wait import wait_random_exponential

from ice.cache import diskcache
from ice.settings import settings
from ice.trace import add_fields
from ice.trace import trace

log = get_logger()


class RateLimitError(Exception):
    def __init__(self, response: httpx.Response):
        self.response = response
        try:
            message = response.json()["error"]["message"]
        except Exception:
            message = response.text[:100]
        super().__init__(message)


def log_attempt_number(retry_state):
    if retry_state.attempt_number > 1:
        exception = retry_state.outcome.exception()
        exception_name = exception.__class__.__name__
        exception_message = str(exception)
        OPENAI_ORG_ID = settings.OPENAI_ORG_ID
        log.warning(
            f"Retrying ({exception_name}: {exception_message}): "
            f"Attempt #{retry_state.attempt_number} ({OPENAI_ORG_ID = })..."
        )


def make_headers() -> dict[str, str]:
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {settings.OPENAI_API_KEY}",
    }
    if settings.OPENAI_ORG_ID:
        headers["OpenAI-Organization"] = settings.OPENAI_ORG_ID
    return headers


RETRYABLE_STATUS_CODES = {408, 429, 502, 503, 504}
OPENAI_BASE_URL = "https://api.openai.com/v1"


def is_retryable_HttpError(e: BaseException) -> bool:
    return (
        isinstance(e, httpx.HTTPStatusError)
        and e.response.status_code in RETRYABLE_STATUS_CODES
    )


class TooLongRequestError(ValueError):
    def __init__(self, prompt: str = "", detail: str = ""):
        self.prompt = prompt
        self.detail = detail
        super().__init__(self.detail)


def raise_if_too_long_error(prompt: object, response: Response) -> None:
    # Raise something more specific than
    # a generic status error if we have exceeded
    # an OpenAI model's context window
    if not isinstance(prompt, str) or response.status_code != 400:
        return None
    try:
        body = response.json()
    except Exception:
        return None
    if not isinstance(body, dict):
        return None
    message = body.get("error", dict).get("message", "")
    if not isinstance(message, str):
        return None
    # This is a bit fragile, but since OpenAI can
    # return 400s for other reasons, checking
    # the message seems like the only real
    # way to tell.
    if "maximum context length" not in message:
        return None
    raise TooLongRequestError(prompt=prompt, detail=message)


@diskcache()
@retry(
    retry=retry_any(
        retry_if_exception(is_retryable_HttpError),
        retry_if_exception_type(TimeoutException),
        retry_if_exception_type(RateLimitError),
    ),
    wait=wait_random_exponential(min=1),
    after=log_attempt_number,
)
async def _post(
    endpoint: str, json: dict, timeout: Optional[float] = None, cache_id: int = 0
) -> Union[dict, TooLongRequestError]:
    """Send a POST request to the OpenAI API and return the JSON response."""
    cache_id  # unused

    async with httpx.AsyncClient() as client:
        response = await client.post(
            f"{OPENAI_BASE_URL}/{endpoint}",
            json=json,
            headers=make_headers(),
            timeout=timeout or 60,
        )
        if response.status_code == 429:
            raise RateLimitError(response)
        try:
            raise_if_too_long_error(prompt=json.get("prompt"), response=response)
        except TooLongRequestError as tlre:
            # Hack alert: Don't raise here so this gets cached
            return tlre
        response.raise_for_status()
        return response.json()


# TODO: support more model types for conversion


def get_davinci_equivalent_tokens(response: dict) -> int:
    return response.get("usage", {}).get("total_tokens", 0)


@trace
async def openai_complete(
    prompt: str,
    stop: Optional[str] = "\n",
    top_p: float = 1,
    temperature: float = 0,
    model: str = "text-davinci-002",
    max_tokens: int = 256,
    logprobs: Optional[int] = None,
    logit_bias: Optional[Mapping[str, Union[int, float]]] = None,
    n: int = 1,
    echo: bool = False,
    cache_id: int = 0,  # for repeated non-deterministic sampling using caching
) -> dict:
    """Send a completion request to the OpenAI API and return the JSON response."""
    params = {
        "prompt": prompt,
        "stop": stop,
        "top_p": top_p,
        "temperature": temperature,
        "model": model,
        "echo": echo,
        "max_tokens": max_tokens,
        "logprobs": logprobs,
        "n": n,
    }
    if logit_bias:
        params["logit_bias"] = logit_bias  # type: ignore[assignment]
    response = await _post("completions", json=params, cache_id=cache_id)
    if isinstance(response, TooLongRequestError):
        raise response
    add_fields(davinci_equivalent_tokens=get_davinci_equivalent_tokens(response))
    return response



from typing import Optional

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import list_experiments
from ice.paper import get_full_document_id
from ice.paper import Paper
from ice.recipe import Recipe
from ice.recipes.evaluate_result import EvaluateResult
from ice.trace import recorder
from ice.utils import truncate_by_tokens

QUESTION_SHORT_NAME = "placebo"

DEFAULT_ANSWER_CLASSIFICATION = "Placebo"


def get_paper_text(paper: Paper) -> str:
    return "\n\n".join(str(p) for p in paper.paragraphs)


def create_recipe_result(paper_id: str, experiment: str, answer: str) -> RecipeResult:
    return RecipeResult(
        document_id=paper_id,
        question_short_name=QUESTION_SHORT_NAME,
        experiment=experiment,
        classifications=[
            DEFAULT_ANSWER_CLASSIFICATION,
            DEFAULT_ANSWER_CLASSIFICATION,
        ],
        answer=answer,
        result=answer,
        excerpts=[],
    )


def get_gold_placebo(paper_id: str, experiment: str) -> Optional[str]:
    """
    Return the gold standard placebo description for the given paper and experiment.
    """
    gold_standard = get_gold_standard(
        document_id=paper_id, question_short_name="placebo", experiment=experiment
    )
    if gold_standard is None:
        return None
    return gold_standard.answer


class PlaceboDescription(Recipe):
    agent_str: str

    def make_prompt(self, paper: Paper, experiment: str) -> str:
        raise NotImplementedError()

    async def get_gold_experiments(self, paper: Paper) -> list[str]:
        paper_id = get_full_document_id(paper.document_id)
        experiments: list[str] = list_experiments(document_id=paper_id)
        return experiments

    async def placebo_for_experiment(
        self, paper: Paper, experiment: str, record=recorder
    ) -> str:

        # Generate the QA prompt
        qa_prompt = self.make_prompt(paper, experiment)

        # Ask the agent to answer the prompt
        placebo_description = await self.agent(self.agent_str).complete(
            prompt=qa_prompt, max_tokens=300
        )

        # Save to trace
        evaluation = await EvaluateResult(mode=self.mode).run(
            gold_result=get_gold_placebo(paper.document_id, experiment),
            model_result=placebo_description,
            question=f'What was the placebo for the "{experiment}" experiment?',
        )
        record(evaluation=evaluation)

        return placebo_description

    async def run(self, paper: Paper, record=recorder):
        # Get the list of experiments (for now from gold standards)
        experiments = await self.get_gold_experiments(paper)

        results = {}
        for experiment in experiments:
            gold_placebo = get_gold_placebo(paper.document_id, experiment)
            if not gold_placebo or gold_placebo in (
                "No placebo",
                "Not mentioned",
                "Unclear",
            ):
                continue

            placebo = await self.placebo_for_experiment(paper, experiment)

            # Save results for each experiment
            recipe_results = [
                create_recipe_result(paper.document_id, experiment, placebo)
            ]
            self.maybe_add_to_results(recipe_results)

            results[experiment] = placebo

        return results


class PlaceboDescriptionInstruct(PlaceboDescription):
    name = "PlaceboDescriptionInstruct"
    agent_str = "instruct"

    def make_prompt(self, paper: Paper, experiment: str) -> str:
        paper_text = truncate_by_tokens(str(paper), max_tokens=3500)
        return f"""Text of a research paper:

{paper_text}

Question: In this paper, what was the placebo in the "{experiment}" experiment?

Answer:"""



from ice.recipe import recipe
from ice.recipes.abstract_qa import Abstract
from ice.recipes.abstract_qa import abstract_qa
from ice.recipes.abstract_qa import DEFAULT_ABSTRACTS
from ice.recipes.combine_abstract_answers import combine_abstract_answers
from ice.utils import map_async


async def synthesize_compositional(abstracts: list[Abstract], question: str) -> str:
    answers = await map_async(
        abstracts, lambda abstract: abstract_qa(abstract=abstract, question=question)
    )
    answer = await combine_abstract_answers(
        question=question, abstracts=abstracts, answers=answers
    )
    return answer


async def synthesize_compositional_cli() -> str:
    abstracts = DEFAULT_ABSTRACTS
    question = "what is the relationship between income and smoking?"
    answer = await synthesize_compositional(abstracts=abstracts, question=question)
    return answer


recipe.main(synthesize_compositional_cli)



from fvalues import F

from ice.recipe import recipe


def make_computation_choice_prompt(question: str) -> str:
    return F(
        f"""You've been asked to answer the question "{question}".

You have access to a Python interpreter.

Enter an expression that will help you answer the question.
>>>"""
    )


def make_compute_qa_prompt(question: str, expression: str, result: str) -> str:
    return F(
        f"""A recording of a Python interpreter session:

>>> {expression}: {result}

Answer the following question, using the Python session if helpful:

Question: "{question}"
Answer: "
"""
    ).strip()


def eval_python(expression: str) -> str:
    try:
        result = eval(expression)
    except Exception as e:
        result = F(f"Error: {e}")
    return str(result)


async def choose_computation(question: str) -> str:
    prompt = make_computation_choice_prompt(question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


async def answer_by_computation(question: str):
    expression = await choose_computation(question)
    result = eval_python(expression)
    prompt = make_compute_qa_prompt(question, expression, result)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(answer_by_computation)



import re

from collections.abc import Sequence

from numerizer import numerize
from structlog.stdlib import get_logger

log = get_logger()


def _extract_nums(text: str) -> Sequence[str]:
    numerized = numerize(text)
    # TODO: handle commas, decimals
    return re.findall("[0-9]+", numerized)


def extract_nums(text: str) -> list[int]:
    return [int(num) for num in _extract_nums(text)]


def strip_enumeration_prefix(text: str) -> str:
    return re.sub(r"^\w*\s*\d+(\.|\))", "", text.strip()).strip()



from typing import Optional

import httpx

from structlog import get_logger

from ice.recipe import recipe
from ice.recipes.elicit.common import send_elicit_request

log = get_logger()


def make_request_body(
    query: str, num_papers: int = 4, page: int = 0, filters: Optional[dict] = None
) -> dict:
    """
    Make the request body for the Elicit search endpoint.
    """
    if filters is None:
        filters = {}
    return dict(
        query=query,
        start=page * num_papers,
        stop=(page + 1) * num_papers,
        qaColumns=[],
        filters=filters,
    )


async def get_elicit_backend() -> str:
    BACKEND_URL = "https://elicit.org/api/backend"
    async with httpx.AsyncClient() as client:
        response = await client.get(BACKEND_URL)
        # Response is plain text, e.g. "https://prod.elicit.org/elicit-red/lit-review"
        response.raise_for_status()
        return response.text


async def elicit_search(
    question: str = "What is the effect of creatine on cognition?",
    num_papers: int = 4,
    page: int = 0,
    has_pdf_filter: bool = False,
    backend: Optional[str] = None,
    filters: Optional[dict] = None,
):
    """
    Search Elicit for papers related to a question.
    """

    backend = backend or await get_elicit_backend()

    endpoint = backend.rstrip("/") + "/lit-review"

    log.info(f"Searching Elicit for query: {question}, endpoint: {endpoint}")

    filters = filters or {}
    if has_pdf_filter:
        filters["has_pdf"] = True

    request_body = make_request_body(
        query=question, num_papers=num_papers, page=page, filters=filters
    )

    response = send_elicit_request(
        request_body=request_body,
        endpoint=endpoint,
    )
    return response


recipe.main(elicit_search)



from fvalues import F

from ice.recipe import recipe


def make_computation_choice_prompt(question: str) -> str:
    return F(
        f"""You've been asked to answer the question "{question}".

You have access to a Python interpreter.

Enter an expression that will help you answer the question.
>>>"""
    )


def eval_python(expression: str) -> str:
    try:
        result = eval(expression)
    except Exception as e:
        result = F(f"Error: {e}")
    return str(result)


async def choose_computation(question: str) -> str:
    prompt = make_computation_choice_prompt(question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


async def eval_selective(question: str):
    expression = await choose_computation(question)
    result = eval_python(expression)
    return (expression, result)


recipe.main(eval_selective)



from fvalues import F

from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import recipe


def make_prompt(paragraph: Paragraph, question: str) -> str:
    return F(
        f"""
Here is a paragraph from a research paper: "{paragraph}"

Question: Does this paragraph answer the question '{question}'? Say Yes or No.
Answer:"""
    ).strip()


async def classify_paragraph(paragraph: Paragraph, question: str) -> float:
    choice_probs, _ = await recipe.agent().classify(
        prompt=make_prompt(paragraph, question),
        choices=(" Yes", " No"),
    )
    return choice_probs.get(" Yes", 0.0)


async def answer_for_paper(paper: Paper, question: str):
    paragraph = paper.paragraphs[0]
    return await classify_paragraph(paragraph, question)


recipe.main(answer_for_paper)



import dataclasses

from inspect import isfunction
from math import isnan
from typing import Any
from typing import Union

from fvalues import F


JSONValue = Union[
    str, int, float, bool, None, list["JSONValue"], dict[str, "JSONValue"]
]


def to_json_value(x: Any) -> JSONValue:
    if isinstance(x, dict):
        return {
            k if isinstance(k, str) else repr(k): to_json_value(v) for k, v in x.items()
        }
    if isinstance(x, (list, tuple, set)):
        return [to_json_value(v) for v in x]
    if isinstance(x, F):
        return {"__fstring__": to_json_value(x.parts)}
    if hasattr(x, "dict") and callable(x.dict):
        try:
            x = x.dict()
        except TypeError:  # raised if wrong number of arguments
            pass
        else:
            return to_json_value(x)
    if dataclasses.is_dataclass(x):
        return to_json_value(dataclasses.asdict(x))
    if isfunction(x):
        return dict(class_name=x.__class__.__name__, name=x.__name__)
    if isinstance(x, float):
        return None if isnan(x) else x
    if isinstance(x, (int, str, bool, type(None))):
        return x
    return repr(x)



from collections.abc import Sequence
from collections.abc import Sized
from typing import Optional
from typing import Union

from structlog.stdlib import get_logger

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.multi import StopSentinel
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.dependent import plural_transform
from ice.formatter.transform.positional import OrdinalWord
from ice.formatter.transform.value import numbered_list
from ice.formatter.transform.value import ValueTransform
from ice.recipes.experiments_and_arms.num_utils import extract_nums
from ice.recipes.experiments_and_arms.prompts.utils import get_part
from ice.recipes.experiments_and_arms.types import MultipartReasoningPrompt

log = get_logger()

COUNT_EXPERIMENTS_PREFACE = """When evaluating a randomized controlled trial, you should first identify the number of distinct experiments (e.g., randomized controlled trials, or RCTs) conducted. An experiment is where the authors evaluate the effect of some program, action, or substance on a population. Two experiments are distinct if they are conducted on different populations. Experiments have different "trial arms". Trial arms describe the different treatment conditions participants were in.

Let's look at excerpts from five different papers to see what information, if any, they provide about the study's experiments (e.g., RCTs). We'll have to identify what each extract tells us about whether the authors conducted one or several experiments. If there is no explicit information indicating that multiple experiments or studies were conducted, assume there was just one experiment.""".strip()

COUNT_EXPERIMENTS_EXAMPLE_TEMPLATE = """
{ordinal}, consider the following excerpt{maybe_plural_excerpts} from a paper:

{paragraphs}

Let's think about what each excerpt tells us, if anything, about the number of experiments (e.g., RCTs): {reasoning}

Here's all the information in this paper about how many experiments there were: {answer}
""".strip()

COUNT_EXPERIMENTS_EXAMPLES: list[
    dict[str, Union[ValueTransform[Sequence[str]], str]]
] = [
    dict(
        paragraphs=numbered_list(
            [
                """Table 3 shows the various comparisons that were made between the treatment groups. To examine the effect of iron supplementation, groups 3 and 5 were comp~red with groups 2 and 4, as the treatments given to those in the former groups were identical to the treatments given to those in the latter groups apart from the addition of iron to groups 3 and 5 (Table 2 ). Similarly the effect of folic acid was examined by comparing groups 2 and 3 with groups 4 and 5. The third comparison in Table 3 was made to determine whether any interaction between iron and folate effects existed. The effect of antimalarials was examined using two comparisons. First, group 1 was compared with group 2, the treatments given to patients in these groups being identical apart from the addition of antimalarials to group 2. Secondly, patients in group 1 were compared with those in groups 2 to 5 combined: this was done only when there was no marked effect of iron or folic acid on the variable being examined.""",
                """To test each comparison for statistical significance, t-tests were performed for continuous variables, and chi-squared tests for binary variables (with a continuity correction). For binary variables for which any expected frequency was less than 5, Fisher's exact test (2-sided) was performed.""",
                """Logarithmic transformations were made for SFA, RCF, serum B 12 and reticulocyte counts before analysis, as the distributions of these variables were skewed, but they have been re-transformed to the original scale for presentation.""",
                """There were no significant differences in any haematological measurement between the five groups when the patients were entered into the trial (Table 4 ). Ninety-one patients (45•5%) were anaemic. Four patients only had totally normal red cells on the blood film; most commonly seen were moderate ( + +) anisocytosis, with mild ( +) macrocytosis, microcytosis and polychromasia, as described previously (Fleming et al., 1984). Six subjects had elliptocytosis not associated with anaemia. Fifty-one (25•5%) had sickle cell trait, with nine to 12 individuals in each of the five groups: two had Hb-AC (groups 1 and 5). Malaria, predominantly P.falciparum, was observed in 53 (26•5%) of subjects. There was only one significant difference in frequency of parasite density between the treatment groups; P.falciparum gametocytes were more frequent in groups 3 and 5 (to receive iron supplements) than in groups 2 and 4 (not to receive iron supplements) (Table 5) Mean±s.D. t' "' two (60•4%) with malaria were anaemic, compared to 59 (40•1 %) without parasitaemia""",
            ]
        ),
        reasoning="""Excerpt 1 discusses various comparisons made between treatment groups. Because these groups are being compared directly, this indicates that they are part of the same study. Excerpt two just explains the statistical significance test use. Excerpt 3 indicates the data transformations used for variables. Exceprt 4 discusses different haematological measurements take in the study, and how frequently different conditions were observed in the different groups.""",
        answer="""Because there is no explicit indication otherwise, these excerpts indicate that the study conducted one experiment. Final Answer: 1.""",
    ),
    dict(
        paragraphs=numbered_list(
            [
                """Table 3 shows the various comparisons that were made between the treatment groups. To examine the effect of iron supplementation, groups 3 and 5 were comp~red with groups 2 and 4, as the treatments given to those in the former groups were identical to the treatments given to those in the latter groups apart from the addition of iron to groups 3 and 5 (Table 2 ). Similarly the effect of folic acid was examined by comparing groups 2 and 3 with groups 4 and 5. The third comparison in Table 3 was made to determine whether any interaction between iron and folate effects existed. The effect of antimalarials was examined using two comparisons. First, group 1 was compared with group 2, the treatments given to patients in these groups being identical apart from the addition of antimalarials to group 2. Secondly, patients in group 1 were compared with those in groups 2 to 5 combined: this was done only when there was no marked effect of iron or folic acid on the variable being examined.""",
                """To test each comparison for statistical significance, t-tests were performed for continuous variables, and chi-squared tests for binary variables (with a continuity correction). For binary variables for which any expected frequency was less than 5, Fisher's exact test (2-sided) was performed.""",
                """Logarithmic transformations were made for SFA, RCF, serum B 12 and reticulocyte counts before analysis, as the distributions of these variables were skewed, but they have been re-transformed to the original scale for presentation.""",
                """There were no significant differences in any haematological measurement between the five groups when the patients were entered into the trial (Table 4 ). Ninety-one patients (45•5%) were anaemic. Four patients only had totally normal red cells on the blood film; most commonly seen were moderate ( + +) anisocytosis, with mild ( +) macrocytosis, microcytosis and polychromasia, as described previously (Fleming et al., 1984). Six subjects had elliptocytosis not associated with anaemia. Fifty-one (25•5%) had sickle cell trait, with nine to 12 individuals in each of the five groups: two had Hb-AC (groups 1 and 5). Malaria, predominantly P.falciparum, was observed in 53 (26•5%) of subjects. There was only one significant difference in frequency of parasite density between the treatment groups; P.falciparum gametocytes were more frequent in groups 3 and 5 (to receive iron supplements) than in groups 2 and 4 (not to receive iron supplements) (Table 5) Mean±s.D. t' "' two (60•4%) with malaria were anaemic, compared to 59 (40•1 %) without parasitaemia""",
            ]
        ),
        reasoning="""Excerpt 1 discusses various comparisons made between treatment groups. Because these groups are being compared directly, this indicates that they are part of the same study. Excerpt two just explains the statistical significance test use. Excerpt 3 indicates the data transformations used for variables. Exceprt 4 discusses different haematological measurements take in the study, and how frequently different conditions were observed in the different groups.""",
        answer="""The excerpts provided indicate that the study conducted two experiments. Final Answer: 2.""",
    ),
]

COUNT_EXPERIMENTS_SHARED: dict[
    str, Union[OrdinalWord, DependentTransform[Union[int, Sized]]]
] = dict(
    ordinal=OrdinalWord(capitalize=True),
    maybe_plural_excerpts=plural_transform(
        key="paragraphs", singular_case="", plural_case="s"
    ),
)


def count_from_answer(answer: str) -> int:
    nums = extract_nums(answer)
    if not nums:
        log.warning("No number in final answer; expected count", answer=answer)
    return nums[-1] if nums else 0


COUNT_EXPERIMENTS_REASONING_STOP = ("\n\nHere's all",)


def make_count_experiments_prompt_func(num_shots: int) -> MultipartReasoningPrompt:
    def count_experiments_prompt(
        paragraphs: Sequence[str],
        helpfulness: Optional[str] = None,
        reasoning: Optional[str] = None,
    ) -> str:
        helpfulness  # ignored
        last_example: dict[
            str, Union[ValueTransform[Sequence[str]], str, StopSentinel]
        ] = dict(
            paragraphs=numbered_list(paragraphs),
            reasoning=reasoning if reasoning else stop("Excerpt 1"),
        )
        if reasoning:
            last_example["answer"] = stop("")
        shots = format_multi(
            COUNT_EXPERIMENTS_EXAMPLE_TEMPLATE,
            COUNT_EXPERIMENTS_EXAMPLES[:num_shots] + [last_example],
            COUNT_EXPERIMENTS_SHARED,
        )
        return "\n\n".join([COUNT_EXPERIMENTS_PREFACE] + list(shots))

    return count_experiments_prompt


def get_count_exps_reasoning(response: str) -> str:
    return "".join(
        ("Excerpt 1 ", get_part(response, "RCTs): ", "\nHere's all the information"))
    )



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Generic
from typing import Optional
from typing import Protocol
from typing import TypeVar

from ice.paper import Paper
from ice.recipes.meta.eval_text_classification import BinaryClassificationMetrics

AnswerType_contra = TypeVar("AnswerType_contra", contravariant=True)


@dataclass
class PaperQaGoldStandard(Generic[AnswerType_contra]):
    paper: Paper
    question: str
    gold_answer: AnswerType_contra
    short_gold_answer: AnswerType_contra
    gold_support: Sequence[str]


@dataclass
class PaperQaAnswer(Generic[AnswerType_contra]):
    answer: AnswerType_contra
    support_candidates: Sequence[str]
    support_labels: Sequence[bool]
    support_scores: Optional[Sequence[float]] = None


@dataclass
class SequenceGenerationEvaluation(Generic[AnswerType_contra]):
    document_id: str
    question: str
    correct: bool
    detail: str
    metrics: BinaryClassificationMetrics
    gold_answer: AnswerType_contra
    generated_answer: AnswerType_contra
    support: Sequence[str]
    generation_f1_score: float

    def as_dict(self):
        return dict(
            document_id=self.document_id,
            question=self.question,
            correct=self.correct,
            detail=self.detail,
            metrics=self.metrics.as_dict(),
            gold_answer=repr(self.gold_answer),
            generated_answer=repr(self.generated_answer),
            support=self.support,
            f1_score=self.generation_f1_score,
        )


class PaperQaMethod(Protocol[AnswerType_contra]):
    async def __call__(
        self,
        __paper: Paper,
        __question: str,
        __gold_support: Optional[Sequence[str]] = None,
    ) -> PaperQaAnswer[AnswerType_contra]:
        ...


class AnswerEvalMethod(Protocol[AnswerType_contra]):
    async def __call__(
        self,
        question: str,
        ground_truth: AnswerType_contra,
        prediction: AnswerType_contra,
    ) -> tuple[bool, str]:
        ...


class ClassificationEvalMethod(Protocol):
    async def __call__(
        self,
        candidates: Sequence[str],
        predictions: Sequence[bool],
        ground_truth: Sequence[str],
        scores: Optional[Sequence[float]] = None,
    ) -> BinaryClassificationMetrics:
        ...



from collections.abc import Sequence

from more_itertools import circular_shifts

from ice.formatter.multi import format_multi
from ice.recipe import recipe
from ice.recipes.program_search.nodes.answer.types import Demonstration
from ice.recipes.program_search.nodes.answer.types import DemonstrationWithReasoning


INSTRUCTIONS = "My friend came up with the following correct answers for each question but didn't write out his reasoning that supports each answer. Can you write a paragraph explaining why each answer is correct? Be sure to quote the parts of the text that support the answer, explaining why this is the correct conclusion."

DEMONSTRATION_EXAMPLE = """

Question: {question}

Text that contains the answer:

{texts}

Correct answer to the question: {answer}

""".strip()

PRE_GENERATION = """

Starting with the first question, {question}, provide a paragraph of reasoning that supports the answer based on the following excerpts, being sure to quote the parts of the excerpts that support the answer and explaining why they support the answer. Only include information that is relevant to answering the question.

{texts}

Let's think it over: First, Excerpt 1 says that""".strip()


def make_reasoning_prompt(demonstrations: Sequence[Demonstration]) -> str:
    examples = format_multi(
        DEMONSTRATION_EXAMPLE, [d.as_dict() for d in demonstrations]
    )
    first_example = demonstrations[0].as_dict()
    return "\n\n".join(
        [
            INSTRUCTIONS,
            "\n\n---\n\n".join(examples),
            PRE_GENERATION.format(
                question=first_example["question"],
                texts=first_example["texts"].transform(),
            ),
        ]
    )


async def add_reasoning(
    demonstrations: Sequence[Demonstration],
) -> Sequence[DemonstrationWithReasoning]:
    shifts = circular_shifts(demonstrations)
    prompts = [make_reasoning_prompt(shift) for shift in shifts]

    async def answer_for_prompt(prompt: str):
        return await recipe.agent().complete(prompt=prompt, stop="\n\n")

    completions = [(await answer_for_prompt(prompt)) for prompt in prompts]

    return [
        DemonstrationWithReasoning(
            reasoning=reasoning, question=d.question, texts=d.texts, answer=d.answer
        )
        for reasoning, d in zip(completions, demonstrations)
    ]


recipe.main(add_reasoning)



from collections.abc import Sequence
from typing import cast

from ice.apis.openai import openai_complete
from ice.paper import Paper
from ice.recipe import recipe
from ice.recipes.experiments_and_arms.num_utils import strip_enumeration_prefix
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    CAN_WE_NAME_ARMS_BEST_CHOICE,
)
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    CAN_WE_NAME_ARMS_CHOICES,
)
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    CAN_WE_NAME_ARMS_REASONING_STOP,
)
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    get_can_we_name_arms_helpfulness,
)
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    get_can_we_name_arms_reasoning,
)
from ice.recipes.experiments_and_arms.prompts.can_name_arms import (
    make_can_we_name_arms_prompt,
)
from ice.recipes.experiments_and_arms.prompts.name_arms import get_name_arms_reasoning
from ice.recipes.experiments_and_arms.prompts.name_arms import make_name_arms_from_exps
from ice.recipes.experiments_and_arms.prompts.name_arms import NAME_ARMS_REASONING_STOP
from ice.recipes.experiments_and_arms.prompts.passages_to_keep import (
    keep_most_helpful_paragraphs,
)
from ice.recipes.experiments_and_arms.prompts.quick_list import make_quick_list_prompt
from ice.recipes.experiments_and_arms.recipes.best_passages import (
    rate_helpfulness_with_reasoning,
)
from ice.recipes.experiments_and_arms.recipes.consensus import best_answer_by_consensus
from ice.recipes.experiments_and_arms.recipes.reason_select_and_answer import (
    answer_with_best_reasoning,
)
from ice.recipes.experiments_and_arms.types import PassageWithReasoning
from ice.trace import Recorder
from ice.trace import recorder
from ice.trace import trace


def make_reduce_to_best_answer(experiment_in_question: str):
    async def reduce_to_best_answer(
        candidates: Sequence[PassageWithReasoning[str]],
    ) -> PassageWithReasoning[str]:
        answers = [c.final_answer for c in candidates if c.final_answer]
        question = f"""What were the trial arms (subgroups of participants) conducted in the experiment: {experiment_in_question}"""
        return PassageWithReasoning(
            passage=candidates[0].passage,
            reasoning="",
            helpfulness="",
            final_answer=await (
                # best_answer_by_clustering(question=question, candidates=answers)
                best_answer_by_consensus(question=question, candidates=answers)
            ),
        )

    return reduce_to_best_answer


@trace
async def name_arms(
    paper: Paper,
    experiments: Sequence[str],
    experiment_in_question: str,
    record: Recorder = recorder,
) -> Sequence[str]:
    """What were  the trial arms for this experiment?

    Args:
        paper (Paper): The paper in question.
        experiments (Sequence[str]): All the experiments in the paper.
        experiment_in_question (str): The experiment to identify the trial arms for.
        record (Recorder, optional): (recorder for tracing). Defaults to recorder.

    Returns:
        Sequence[str]: The trial arms for the `experiment_in_question`
    """
    paragraphs = paper.nonempty_paragraphs()
    passages_by_relevance = await rate_helpfulness_with_reasoning(
        [str(p) for p in paragraphs],
        make_can_we_name_arms_prompt(
            experiments=experiments, experiment_in_question=experiment_in_question
        ),
        CAN_WE_NAME_ARMS_CHOICES,
        CAN_WE_NAME_ARMS_BEST_CHOICE,
        reasoning_stop=CAN_WE_NAME_ARMS_REASONING_STOP,
        get_reasoning=get_can_we_name_arms_reasoning,
        get_helpfulness=get_can_we_name_arms_helpfulness,
        num_shots=3,
        passages_per_prompt=4,
        step=1,
    )

    paragraphs_to_keep = await keep_most_helpful_paragraphs(passages_by_relevance)

    arm_names = await answer_with_best_reasoning(
        num_samples=6,  # TODO: Better sampling here
        selector=make_reduce_to_best_answer(experiment_in_question),
        # selector=first,
        texts=paragraphs_to_keep,
        # texts=count_paras,
        num_examples=2,
        reasoning_temperature=0.4,
        reasoning_stop=NAME_ARMS_REASONING_STOP,
        prompt_func=make_name_arms_from_exps(
            experiments=experiments, experiment_in_question=experiment_in_question
        ),
        get_reasoning=get_name_arms_reasoning,
        get_helpfulness=None,
        final_answer_processor=lambda resp: cast(str, resp["choices"][0]["text"]),
    )

    assert arm_names.final_answer is not None

    final_answer: str = await convert_answer_to_standardized_format(
        arm_names.final_answer
    )

    return (
        [
            strip_enumeration_prefix(exp_name)
            for exp_name in final_answer.split("\n")
            if exp_name.strip()
        ]
        if final_answer
        else []
    )


@trace
async def convert_answer_to_standardized_format(answer: str) -> str:
    standardized_answer: str = (
        await openai_complete(
            prompt=make_quick_list_prompt(answer),
            stop="\n\nAnswer:",
        )
    )["choices"][0]["text"]
    return standardized_answer


recipe.main(name_arms)



import re

from ice.evaluation.evaluate_recipe_result import RecipeResult
from ice.metrics.gold_standards import list_experiments
from ice.paper import Paper
from ice.recipe import Recipe

SEARCH_TERMS = ["adhere", "take-up", "compliance", "attrition"]


class AdherenceKeywordBaseline(Recipe):
    @staticmethod
    def answer_adherence_using_keywords(paper: Paper) -> str:
        for term in SEARCH_TERMS:
            for paragraph in paper.paragraphs:
                paragraph_text = " ".join(paragraph.sentences)
                if re.search(term, paragraph_text):
                    return paragraph_text
        return ""

    async def run(self, paper: Paper):
        experiments = list_experiments(document_id=paper.document_id)

        results: list[RecipeResult] = []
        for experiment in experiments:
            answer = self.answer_adherence_using_keywords(paper)
            results.append(
                RecipeResult(
                    experiment=experiment,
                    question_short_name="adherence",
                    document_id=paper.document_id,
                    answer=answer,
                    classifications=[None, None],
                    excerpts=[answer],
                )
            )

        self.maybe_add_to_results(results)

        return results



from collections.abc import Sequence
from typing import ClassVar
from typing import Literal
from typing import Optional
from typing import Type
from typing import Union

from pydantic import BaseModel
from pydantic import root_validator
from pydantic import validator

from ice.metrics.gold_standards import ParsedGoldStandardBase

NOT_YET_COLLECTED = "Gold standard not yet collected"


def _fix_keys(model_values: object, correct_key_mapping: dict[str, str]) -> object:
    if isinstance(model_values, dict):
        new_dict = dict()
        for key, value in model_values.items():
            if key in correct_key_mapping:
                new_dict[correct_key_mapping[key]] = value
            else:
                new_dict[key] = value
        return new_dict
    return model_values


def _not_yet_collected_is_none(value: object) -> Optional[object]:
    if value == NOT_YET_COLLECTED:
        return None
    return value


def _maybe_dict_to_sequence(
    constructor: Type[BaseModel], maybe_dict: object
) -> Union[object, Sequence[BaseModel]]:
    if isinstance(maybe_dict, dict):
        return [
            constructor.parse_obj(vals | dict(name=k)) for k, vals in maybe_dict.items()
        ]
    return maybe_dict


class SampleSize(BaseModel):
    description: Optional[str] = None
    reasoning: Optional[str] = None
    n: Union[int, Literal["Not mentioned"]]
    quotes: Sequence[str]


class Reason(BaseModel):
    name: str
    description: Optional[str] = None
    reasoning: Optional[str] = None
    n: Union[int, Literal["Not mentioned"]]


class SampleSizeWithReasons(SampleSize):
    reasons: Optional[Union[Literal["Not mentioned"], Sequence[Reason]]] = None

    @validator("reasons", pre=True)
    def validate_reasons(cls, v):
        return _maybe_dict_to_sequence(Reason, v)


class Received(BaseModel):
    description: Optional[str] = None
    quotes: Sequence[str]


class Analysis(SampleSize):
    name: str


class Arm(BaseModel):
    name: str
    allocated: Union[Optional[Literal["Not mentioned"]], SampleSize]
    received: Union[Optional[Literal["Not mentioned"]], Received]
    attrition: Optional[SampleSizeWithReasons]
    analyzed: Optional[Sequence[Analysis]]

    @validator("analyzed", pre=True)
    def validate_analyzed(cls, v):
        return _maybe_dict_to_sequence(Analysis, v)

    @root_validator(pre=True)
    def not_yet_collected_always_none(cls, values):
        if isinstance(values, dict):
            for key, value in values.items():
                if value == NOT_YET_COLLECTED:
                    values[key] = None
        return _fix_keys(
            values,
            {
                "Allocated to arm": "allocated",
                "Received allocated intervention": "received",
                "-> Attrition": "attrition",
                "Analysed": "analyzed",
            },
        )


class Enrolment(BaseModel):
    assessed: Union[Literal["Not mentioned"], SampleSize]
    excluded: Union[Literal["Not mentioned"], SampleSizeWithReasons]
    randomized: SampleSize

    @root_validator(pre=True)
    def fix_keys(cls, values):
        return _fix_keys(
            values,
            {
                "Assessed for eligibility": "assessed",
                "-> Excluded": "excluded",
                "Randomised": "randomized",
            },
        )


class Experiment(BaseModel):
    name: str
    description: Optional[str] = None
    reasoning: Optional[str] = None
    enrolment: Optional[Enrolment]
    arms: Optional[Sequence[Arm]]

    _enrolment_to_none = validator("enrolment", pre=True, allow_reuse=True)(
        _not_yet_collected_is_none
    )

    @validator("arms", pre=True)
    def validate_arms(cls, v):
        return _maybe_dict_to_sequence(Arm, v)

    @root_validator(pre=True)
    def fix_keys(cls, values):
        return _fix_keys(
            values, {"Enrolment": "enrolment", "assigning to arms and after": "arms"}
        )


class ConsortFlow(ParsedGoldStandardBase):
    question_short_name: ClassVar[str] = "consort_flow_v2"
    experiments: list[Experiment]

    @validator("experiments", pre=True)
    def validate_experiments(cls, v):
        return _maybe_dict_to_sequence(Experiment, v)



import re

from collections.abc import Callable
from dataclasses import dataclass
from time import perf_counter

import torch
import torch.nn.functional as F

from structlog.stdlib import get_logger
from torch import nn
from transformers.models.auto.modeling_auto import AutoModelForSeq2SeqLM
from transformers.models.auto.tokenization_auto import AutoTokenizer
from transformers.models.t5.modeling_t5 import T5ForConditionalGeneration
from transformers.models.t5.tokenization_t5_fast import T5TokenizerFast

log = get_logger()

LORA_MODULES = ".*SelfAttention|.*EncDecAttention|.*DenseReluDense"
LORA_LAYERS = "k|v|wi_1.*"


@dataclass
class PromptedClassificationInput:
    prompt: str
    choices: tuple[str, ...]


@dataclass
class T5Batch:
    input_ids: torch.LongTensor
    choices_ids: torch.Tensor


@dataclass
class PromptedClassificationOutput:
    prediction: str
    prob: float


def create_collate_fn(
    tokenizer: T5TokenizerFast,
) -> Callable[[list[PromptedClassificationInput]], T5Batch]:
    def collate_fn(items: list[PromptedClassificationInput]) -> T5Batch:
        input_ids: torch.LongTensor = tokenizer.batch_encode_plus(
            [item.prompt for item in items],
            padding=True,
            truncation=True,
            return_tensors="pt",
        )["input_ids"]
        num_choices = [len(item.choices) for item in items]
        flat_choices_ids: torch.LongTensor = tokenizer.batch_encode_plus(
            [choice for item in items for choice in item.choices],
            padding=True,
            truncation=True,
            return_tensors="pt",
        )["input_ids"]
        choices_ids = flat_choices_ids.view(
            len(items), max(num_choices), -1
        ).contiguous()
        return T5Batch(input_ids, choices_ids)

    return collate_fn


class TfewInferenceModel(nn.Module):
    def __init__(
        self,
        origin_model: T5ForConditionalGeneration,
        tokenizer: T5TokenizerFast,
        lora_weights_path: str,
    ):
        super().__init__()
        self.model = origin_model
        self.tokenizer = tokenizer
        self.collate_fn = create_collate_fn(tokenizer)
        lora_weights = torch.load(lora_weights_path, map_location="cpu")
        load_result = self.model.load_state_dict(lora_weights, strict=False)
        assert (
            not load_result.unexpected_keys
        ), f"Unexpected keys found in state dict: {load_result.unexpected_keys}"
        self.model.to(torch.bfloat16)

    def predict(
        self, batch: list[PromptedClassificationInput]
    ) -> list[PromptedClassificationOutput]:
        with torch.no_grad():
            return self._predict(batch)

    def _predict(
        self, batch: list[PromptedClassificationInput]
    ) -> list[PromptedClassificationOutput]:
        start = perf_counter()
        t5_batch = self.collate_fn(batch)
        input_ids, choices_ids = t5_batch.input_ids, t5_batch.choices_ids
        input_ids, choices_ids = input_ids.to(self.model.device), choices_ids.to(
            self.model.device
        )
        bs, num_choices = choices_ids.size()[:2]
        flat_choices_ids = choices_ids.flatten(0, 1)
        attention_mask = (
            input_ids != self.tokenizer.pad_token_id
        ).float()  # [bs, max_seq_len]
        encoder_hidden_states = self.model.encoder(
            input_ids=input_ids, attention_mask=attention_mask
        )[0]
        encoder_hidden_states = (
            encoder_hidden_states.unsqueeze(dim=1)
            .repeat(1, num_choices, 1, 1)
            .flatten(0, 1)
        )
        attention_mask = (
            attention_mask.unsqueeze(dim=1).repeat(1, num_choices, 1).flatten(0, 1)
        )
        decoder_input_ids = torch.cat(
            [torch.zeros_like(flat_choices_ids[:, :1]), flat_choices_ids[:, :-1]], dim=1
        )
        decoder_attention_mask = (decoder_input_ids == decoder_input_ids).float()
        lm_target = (
            flat_choices_ids
            - 100 * (flat_choices_ids == self.tokenizer.pad_token_id).long()
        )

        model_output = self.model(
            attention_mask=attention_mask,
            encoder_outputs=[encoder_hidden_states],
            decoder_input_ids=decoder_input_ids,
            decoder_attention_mask=decoder_attention_mask,
        )
        choices_scores = (
            F.cross_entropy(
                model_output.logits.flatten(0, 1),
                lm_target.flatten(0, 1),
                reduction="none",
            )
            .view(bs, num_choices, -1)
            .sum(dim=-1)
        )
        # Length normalization
        choices_scores = choices_scores / (
            choices_ids != self.tokenizer.pad_token_id
        ).sum(dim=-1)
        _, prediction = choices_scores.min(dim=1)
        probs, _ = F.softmax(choices_scores, dim=1).max(dim=1)
        pred_list = prediction.tolist()
        pred_choices = [
            example.choices[pred_list[i]] for i, example in enumerate(batch)
        ]
        log.info(
            "T-Few batch inference",
            batch_size=len(batch),
            latency_in_s=perf_counter() - start,
        )
        return [
            PromptedClassificationOutput(prediction=pred_choice, prob=prob)
            for pred_choice, prob in zip(pred_choices, probs.tolist())
        ]


class LoRALinear(nn.Module):
    def __init__(
        self,
        linear_layer: nn.Linear,
    ):
        super().__init__()
        self.in_features = linear_layer.in_features
        self.out_features = linear_layer.out_features
        self.weight = linear_layer.weight
        self.bias = linear_layer.bias
        self.multi_lora_a = nn.parameter.Parameter(
            torch.ones(1, linear_layer.in_features)
        )
        self.multi_lora_b = nn.parameter.Parameter(
            torch.ones(linear_layer.out_features, 1)
        )

    def forward(self, input):
        hidden = F.linear((input * self.multi_lora_a.flatten()), self.weight, self.bias)
        hidden = hidden * self.multi_lora_b.flatten()
        return hidden

    def extra_repr(self):
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"


def inject_low_rank_adaptation(transformer: torch.nn.Module):
    for m_name, module in dict(transformer.named_modules()).items():
        if re.fullmatch(LORA_MODULES, m_name):
            for c_name, layer in dict(module.named_children()).items():
                if re.fullmatch(LORA_LAYERS, c_name):
                    assert isinstance(
                        layer, nn.Linear
                    ), f"LoRA can only be applied to torch.nn.Linear, but {layer} is {type(layer)}."
                    setattr(
                        module,
                        c_name,
                        LoRALinear(layer),
                    )
    return transformer


def make_origin_model(
    origin_model_name: str, max_seq_len: int = 1024
) -> tuple[T5TokenizerFast, T5ForConditionalGeneration]:
    tokenizer: T5TokenizerFast = AutoTokenizer.from_pretrained(origin_model_name)
    origin_model: T5ForConditionalGeneration = AutoModelForSeq2SeqLM.from_pretrained(
        origin_model_name
    )
    tokenizer.model_max_length = max_seq_len
    origin_model = inject_low_rank_adaptation(origin_model)
    origin_model.eval()
    if torch.cuda.is_available():
        origin_model.cuda()
    return tokenizer, origin_model


def load_inference_model(
    origin_model_name: str, lora_weights_path: str, max_seq_len: int = 1024
) -> TfewInferenceModel:
    start = perf_counter()
    log.info(
        "Loading tfew model",
        origin_model_name=origin_model_name,
        lora_weights_path=lora_weights_path,
    )
    tokenizer, origin_model = make_origin_model(origin_model_name, max_seq_len)
    model = TfewInferenceModel(origin_model, tokenizer, lora_weights_path)
    log.info(
        "Loaded tfew model", elapsed_s=perf_counter() - start, device=model.model.device
    )
    return model



import enum
import string

from collections import Counter
from collections.abc import Mapping
from collections.abc import Sequence
from typing import final
from typing import Optional
from typing import Union

from structlog.stdlib import get_logger
from typing_extensions import TypeGuard

from ice.formatter.transform import _Transform
from ice.formatter.transform.dependent import DependentTransform
from ice.formatter.transform.positional import PositionalTransform
from ice.formatter.transform.value import ValueTransform

log = get_logger()

# Tools for building few-shot prompts out of examples

_formatter = string.Formatter()


class _NotNeededSentinel(enum.Enum):
    token = 0


_not_needed_sentinel = _NotNeededSentinel.token


@final
class StopSentinel(str):
    """Stop the string here."""


literal = Union[str, int, float, StopSentinel]

literal_or_transform = Union[literal, _Transform]

FormatValues = Sequence[Mapping[str, literal_or_transform]]

_StdLibFormatStringParses = list[
    tuple[str, Optional[str], Optional[str], Optional[str]]
]


def _is_partial(**fields: Union[literal, _NotNeededSentinel]):
    return any((isinstance(value, StopSentinel) for value in fields.values()))


def all_values_needed(
    examples: Sequence[Mapping[str, Union[literal_or_transform, _NotNeededSentinel]]]
) -> TypeGuard[Sequence[Mapping[str, literal_or_transform]]]:
    return all(
        (
            all(v is not _not_needed_sentinel for v in example.values())
            for example in examples
        )
    )


def _is_keyword_only(parses: _StdLibFormatStringParses) -> bool:
    return all(
        [
            field_name is None or field_name.isidentifier()
            for _, field_name, _, _ in parses
        ]
    )


def _parse(format_placeholder: str) -> _StdLibFormatStringParses:
    parses = list(_formatter.parse(format_placeholder))
    if not _is_keyword_only(parses):
        raise ValueError(
            f"placeholder must use only keyword placeholders: {format_placeholder}"
        )
    return parses


def _literal_brace_escape(literal_text: str) -> str:
    chars = Counter(literal_text)
    # For safety, ensuring these came from string.Formatter.parse
    assert chars.get("{", 0) + chars.get("}", 0) <= 1
    if not literal_text:
        return literal_text
    if literal_text[-1] == "}":
        return "".join((literal_text, "}"))
    if literal_text[-1] == "{":
        return "".join((literal_text, "{"))
    return literal_text


def _unparse(parses: _StdLibFormatStringParses) -> str:
    # Suprisingly this isn't in the stdlib
    parts: list[str] = []
    for literal_text, field_name, format_spec, conversion in parses:
        parts.append(_literal_brace_escape(literal_text))
        if field_name is not None:
            parts.extend(["{", field_name])
            if conversion is not None:
                parts.extend(["!", conversion])
            if format_spec is not None:
                parts.extend([":", format_spec])
            parts.append("}")
    return "".join(parts)


def _no_sentinels_remaining(
    concrete_values: dict[str, Union[literal, _NotNeededSentinel]]
) -> TypeGuard[dict[str, literal]]:
    return all(
        (value is not _not_needed_sentinel for value in concrete_values.values())
    )


def _format_truncate(
    format_placeholder: str,
    /,
    **concrete_values: Union[literal, _NotNeededSentinel],
) -> str:
    if not _is_partial(**concrete_values):
        return format_placeholder
    parses = _parse(format_placeholder)
    keep: _StdLibFormatStringParses = []
    for parse in parses:
        _, field_name, _, _ = parse
        if field_name is None:
            keep.append(parse)
        else:
            keep.append(parse)
            if isinstance(concrete_values[field_name], StopSentinel):
                break
    needed_keys = set(
        (field_name for _, field_name, _, _ in keep if field_name is not None)
    )
    needed_concrete_values = {
        k: v for k, v in concrete_values.items() if k in needed_keys
    }
    if not _no_sentinels_remaining(needed_concrete_values):
        raise ValueError(
            f"Unfilled DependentTransform in partial example:\nFormat: {format_placeholder}\nNeeded Values: {needed_concrete_values}"
        )
    new_str = _unparse(keep).format_map(needed_concrete_values)
    return new_str


def _has_stop(
    concrete_values: Mapping[str, Union[literal, _NotNeededSentinel]]
) -> bool:
    return any(isinstance(value, StopSentinel) for value in concrete_values.values())


def _format_single(
    format_placeholder: str,
    /,
    **concrete_values: Union[literal, _NotNeededSentinel],
) -> tuple[str, bool]:
    if not _has_stop(concrete_values):
        formatted = format_placeholder.format_map(concrete_values)
        truncated = False
    else:
        formatted = _format_truncate(format_placeholder, **concrete_values)
        truncated = True
    return formatted, truncated


def _apply_transforms(
    cases: Sequence[Mapping[str, literal_or_transform]],
    total: int,
) -> Sequence[Mapping[str, Union[literal, _NotNeededSentinel]]]:
    ret_list: list[dict[str, Union[literal, _NotNeededSentinel]]] = []
    for position, case in enumerate(cases):
        ret_dict: dict[str, Union[literal, _NotNeededSentinel]] = {}
        for k, v in case.items():
            if not isinstance(v, _Transform):
                ret_dict[k] = v
            elif isinstance(v, ValueTransform):
                ret_dict[k] = v.transform()
            elif isinstance(v, PositionalTransform):
                ret_dict[k] = v.transform(position, total)
            elif isinstance(v, DependentTransform):
                if v.key() not in case:
                    ret_dict[k] = _not_needed_sentinel
                else:
                    dependent = case[v.key()]
                    ret_dict[k] = (
                        v.transform(dependent.value)
                        if isinstance(dependent, ValueTransform)
                        else v.transform(dependent)
                    )
        ret_list.append(ret_dict)
    return ret_list


_STOP_EARLY_WARNING = "StopSentinel used in non-final case; this may be a mistake."


def _format_multi(
    format_string: str,
    cases: Sequence[Mapping[str, literal_or_transform]],
    strip: bool,
) -> tuple[str, ...]:
    total = len(cases)
    transformed_examples = _apply_transforms(cases, total)
    filled_cases = list(
        map(
            lambda kwargs: _format_single(format_string, **kwargs),
            transformed_examples,
        )
    )
    ret_val, truncated = tuple(c[0].strip() if strip else c[0] for c in filled_cases), [
        c[1] for c in filled_cases
    ]
    if any(truncated[:-1]):
        log.warn(_STOP_EARLY_WARNING, format_string=format_string, cases=cases)
    return ret_val


def stop(value: Union[str, int, float]) -> StopSentinel:
    """Stop here; truncate the rest of the format string."""
    return StopSentinel(value)


def format_multi(
    format_str: str,
    cases: Sequence[Mapping[str, literal_or_transform]],
    shared: Optional[Mapping[str, literal_or_transform]] = None,
    strip: bool = True,
) -> tuple[str, ...]:
    """
    str.Formatter on steroids.

    Repeatedly format a string based on the arguments passed in `cases`.
    Shared arguments belong in `shared`.

    Use Transforms to dynamically change the value based on other values,
    the index of the case, the number of cases, or another function.

    Stop early and truncate a case with a StopSentinel value.
    """
    shared_dict = {k: v for k, v in (shared or dict()).items()}
    return _format_multi(
        format_str, [shared_dict | case for case in cases], strip=strip
    )



from collections.abc import Iterator
from collections.abc import Sequence
from functools import cache
from functools import cached_property
from pathlib import Path
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Optional
from typing import overload
from typing import Type
from typing import TypeVar

import pandas as pd

from pydantic import BaseModel
from pydantic.generics import GenericModel
from structlog.stdlib import get_logger
from yaml import CLoader as Loader
from yaml import load

from ice.paper import Paper
from ice.settings import settings

log = get_logger()


class ParsedGoldStandardBase(BaseModel):
    question_short_name: ClassVar[str]


ParsedGoldStandardType = TypeVar("ParsedGoldStandardType", bound=ParsedGoldStandardBase)

GoldStandardSplit = Literal["test", "validation", "iterate"]


class GoldStandard(GenericModel, Generic[ParsedGoldStandardType]):
    document_id: str
    question_short_name: str
    experiment: str
    answer: str
    classifications: Sequence[Optional[str]] = []
    quotes: list[str]
    split: Optional[str] = None
    answer_model: Optional[Type[ParsedGoldStandardType]] = None

    @cached_property
    def parsed_answer(self) -> Optional[ParsedGoldStandardType]:
        return (
            _parse_answer(self.answer, self.answer_model)
            if self.answer_model is not None
            else None
        )

    class Config:
        keep_untouched = (cached_property,)
        fields = dict(answer_model=dict(exclude=True))


def _parse_answer(
    _answer: str, model: Type[ParsedGoldStandardType]
) -> ParsedGoldStandardType:
    data = load(_answer, Loader=Loader)
    return model.parse_obj(data)


@cache
def retrieve_gold_standards_df() -> pd.DataFrame:
    df = pd.read_csv(settings.GOLD_STANDARDS_CSV_PATH)
    df = add_quotes_column(df)
    df = add_classifications_column(df)
    return df


def value_in_column(column: str, value: str, df: pd.DataFrame) -> bool:
    return df[column].str.contains(value, regex=False).any()


def add_quotes_column(df: pd.DataFrame) -> pd.DataFrame:
    quotes = df[[col for col in df.columns if col.startswith("quote_")]]
    df["quotes"] = quotes.apply(
        lambda row: [row[col] for col in quotes.columns if not pd.isna(row[col])],
        axis=1,
    )
    return df


def add_classifications_column(df: pd.DataFrame) -> pd.DataFrame:
    classifications = df[
        [col for col in df.columns if col.startswith("classification_")]
    ]

    ordered_classification_cols = classifications.columns.sort_values()

    df["classifications"] = classifications.apply(
        lambda row: [
            None if pd.isna(row[col]) else row[col]
            for col in ordered_classification_cols
        ],
        axis=1,
    )
    return df


def _standards_df_to_gold_standards(
    df: pd.DataFrame, answer_model: Optional[Type[ParsedGoldStandardType]]
) -> list[GoldStandard[ParsedGoldStandardType]]:
    return [
        GoldStandard.parse_obj(record | dict(answer_model=answer_model))
        for record in df.to_dict("records")
    ]


def list_experiments(
    *, document_id: Optional[str] = None, question_short_name: Optional[str] = None
) -> list[str]:
    df = retrieve_gold_standards_df()
    df = select_column_values(
        df, dict(document_id=document_id, question_short_name=question_short_name)
    )
    experiments = df.experiment.dropna().unique().tolist()
    return [experiment for experiment in experiments if experiment != "All"]


def select_column_values(
    df: pd.DataFrame, column_values: dict[str, Optional[str]]
) -> pd.DataFrame:
    mask = pd.Series(True, index=df.index)
    for column, value in column_values.items():
        if value is not None:
            if not value_in_column(column, value, df):
                log.warning(
                    "Removing all records because filter does not match",
                    column=column,
                    value=value,
                )
            mask &= df[column] == value
    return df.loc[mask]


_paper_dir = settings.PAPER_DIR


def load_paper(document_id: str) -> Paper:
    return Paper.load(Path(_paper_dir, document_id))


def load_papers(split: str, question_short_name: str):
    doc_ids = {
        gs.document_id
        for gs in get_gold_standards(question_short_name=question_short_name)
        if gs.split == split
    }
    return [load_paper(doc_id) for doc_id in doc_ids]


def generate_papers_and_golds(
    split: str,
    gold_standard_type: Type[ParsedGoldStandardType],
) -> Iterator[tuple[Paper, GoldStandard[ParsedGoldStandardType]]]:
    papers = load_papers(split, gold_standard_type.question_short_name)

    for paper in papers:
        gold = get_gold_standard(
            document_id=paper.document_id,
            question_short_name=gold_standard_type.question_short_name,
            model_type=gold_standard_type,
        )
        if not gold:
            log.warning(
                "Did not find gold standard",
                document_id=paper.document_id,
                question_short_name=gold_standard_type.question_short_name,
            )
            continue
        yield paper, gold


@overload
def get_gold_standards(
    *,
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
    model_type: None = None,
) -> list[GoldStandard[Any]]:
    ...


@overload
def get_gold_standards(
    *,
    model_type: Type[ParsedGoldStandardType],
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
) -> list[GoldStandard[ParsedGoldStandardType]]:
    ...


def get_gold_standards(
    *,
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
    model_type: Optional[Type[ParsedGoldStandardType]] = None,
) -> list[GoldStandard[ParsedGoldStandardType]]:
    df = retrieve_gold_standards_df()

    filters = dict(
        document_id=document_id,
        question_short_name=question_short_name,
        experiment=experiment,
    )

    df = select_column_values(df, filters)

    return _standards_df_to_gold_standards(df, model_type)


@overload
def get_gold_standard(
    *,
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
    model_type: None = None,
) -> Optional[GoldStandard[Any]]:
    ...


@overload
def get_gold_standard(
    *,
    model_type: Type[ParsedGoldStandardType],
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
) -> Optional[GoldStandard[ParsedGoldStandardType]]:
    ...


def get_gold_standard(
    *,
    document_id: Optional[str] = None,
    question_short_name: Optional[str] = None,
    experiment: Optional[str] = None,
    model_type: Optional[Type[ParsedGoldStandardType]] = None,
) -> Optional[GoldStandard[ParsedGoldStandardType]]:
    gold_standards = get_gold_standards(
        document_id=document_id,
        question_short_name=question_short_name,
        experiment=experiment,
        model_type=model_type,
    )

    if len(gold_standards) == 0:
        return None

    if len(gold_standards) > 1:
        raise ValueError(
            f"Found more than one gold standard for experiment={experiment}, question_short_name={question_short_name}, document_id={document_id}"
        )

    return gold_standards[0]



from fvalues import F

from ice.recipe import recipe


def make_subquestion_prompt(question: str) -> str:
    return F(
        f"""Decompose the following question into 2-5 subquestions that would help you answer the question. Make the questions stand alone, so that they can be answered without the context of the original question.

Question: "{question}"
Subquestions:
-"""
    ).strip()


async def ask_subquestions(
    question: str = "What is the effect of creatine on cognition?",
):
    prompt = make_subquestion_prompt(question)
    subquestions_text = await recipe.agent().complete(prompt=prompt)
    subquestions = [line.strip("- ") for line in subquestions_text.split("\n")]
    return subquestions


recipe.main(ask_subquestions)



from collections.abc import Sequence
from typing import Optional

from ice.metrics.gold_standards import get_gold_standard
from ice.metrics.gold_standards import GoldStandard
from ice.recipes.experiments_and_arms.types import ExperimentsArms


def get_ea_gs(document_id: str) -> Optional[GoldStandard[ExperimentsArms]]:
    return get_gold_standard(
        document_id=document_id,
        question_short_name="experiments_arms",
        model_type=ExperimentsArms,
    )


def list_gs_experiments(document_id: str) -> Optional[Sequence[str]]:
    gs = get_ea_gs(document_id)
    if gs is None or gs.parsed_answer is None:
        return None
    return [experiment.name for experiment in gs.parsed_answer.experiments]


def count_gs_experiments(document_id: str) -> Optional[int]:
    experiments = list_gs_experiments(document_id)
    return len(experiments) if experiments is not None else None



from fvalues import F

from ice.recipe import recipe


def make_qa_prompt(question: str) -> str:
    return F(
        f"""Answer the following question:

Question: "{question}"
Answer: "
"""
    ).strip()


async def answer(question: str = "What is happening on 9/9/2022?"):
    prompt = make_qa_prompt(question)
    answer = await recipe.agent().complete(prompt=prompt, stop='"')
    return answer


recipe.main(answer)



import re

from collections.abc import Sequence

from ice.formatter.multi import format_multi
from ice.formatter.multi import stop
from ice.formatter.transform.value import numbered_list
from ice.recipe import recipe

PREAMBLE = "For each question, convert the part of the answer that answers the question into a numbered list."

EXAMPLES = [
    dict(
        question="What experiments were conducted in this paper?",
        answer="The six experiments conducted in this paper were 1) an asset index experiment, 2) a consumption support experiment in Ethiopia and Peru, 3) a consumption support experiment in general, 4) an output production experiment, 5) an asset transfer experiment, and 6) a non-farm micro-enterprise experiment.",
        separated=numbered_list(
            [
                "An asset index experiment",
                "A consumption support experiment in Ethiopia and Peru",
                "A consumption support experiment in general",
                "An output production experiment",
                "An asset transfer experiment",
                "A non-farm micro-enterprise experiment",
            ]
        ),
    ),
    dict(
        question="Experiments are distinct from trial arms or groups; a single experiment might have multiple trial arms, like different interventions or controls. What experiment or experiments were conducted in this study? List them out, being mindful that there may just be one experiment or there could be more than one.",
        answer="""1. MIND#1, MIND#2, and MIND#3: three controlled experiments to study the effect of mindfulness on Software Engineering students’ conceptual modeling performance. 2. The experimental group practiced mindfulness while the control group was trained in public speaking as a placebo treatment. 3. The subjects were divided into two groups and all the subjects developed two conceptual models based on a transcript of an interview, one before and another one after the treatment. The results were compared in terms of conceptual modeling quality and productivity. 4. The statistically significant results of the series of experiments revealed that the subjects who practiced mindfulness developed slightly better conceptual models and they did it faster than the control group. 5. The practice of mindfulness improves the performance of Software Engineering students in conceptual modeling, especially their productivity. 6. However, more experimentation is needed in order to confirm the outcomes in other Software Engineering tasks and populations. 7. In the mindfulness workshops, the sessions were face– to–face, four days a week. All the sessions followed the same dynamics: the students and the researcher responsible for conducting the session met in a classroom; they all sat down, lights were turned off and curtains were drawn letting only some dim light in the room; when they all were in silence, an alarm was programmed; during the first five minutes, the subjects were guided in their body scan; then, during the remaining time, they were invited to focus solely on their breathing. Sometimes, the researcher asked “where is your mind now?” in order to re–focus them on breathing. In the event some students were late, they were instructed to enter the room making as less noise as possible and sit on one of the chairs that were intentionally left empty near the door. 8. In the public speaking workshops, the subjects were given some basic guidelines on how to prepare a talk, some notions on non–verbal communication and some seminal talks were commented. Later, they were invited to look for related videos on the Internet and to prepare a script for a public presentation on a topic of their interest. 9. The result of applying changes CH1 to CH3 in MIND#2 resulted in its differential settings consisting of a random assignment of subjects to groups""",
        separated=numbered_list(["MIND#1", "MIND#2", "MIND#3"]),
    ),
    dict(
        question="What were the trial arms of the ivermectin RCT?",
        answer="We conducted two experiments, one on the effectiveness of ivermectin and another on the effectiveness of Paxlovid. Each experiment had two trial arms: a control group and an intervention.",
        separated=numbered_list(["a control group", "an intervention group"]),
    ),
]

TEMPLATE = """
Question: {question}

Answer: {answer}

Convert the answer to a numbered list:

{separated}""".strip()


def make_quick_list_prompt(question: str, answer: str):
    examples = format_multi(
        TEMPLATE, EXAMPLES + [dict(question=question, answer=answer, separated=stop(""))]  # type: ignore[arg-type]
    )
    return "\n\n---\n\n".join([PREAMBLE] + list(examples))


def strip_enumeration_prefix(text: str) -> str:
    return re.sub(r"^\w*\s*\d+(\.|\))", "", text.strip()).strip()


async def quick_list(question: str, answer: str) -> Sequence[str]:
    prompt = make_quick_list_prompt(question, answer)
    completion = await recipe.agent().complete(prompt=prompt, stop="\n\n---")
    return [strip_enumeration_prefix(item) for item in completion.splitlines() if item]


recipe.main(quick_list)



from typing import Optional
from typing import Union

from ice.trace import TracedABC

Stop = Optional[Union[str, list[str]]]


class Agent(TracedABC):
    label: Optional[str] = None

    async def complete(
        self,
        *,
        prompt: str,
        stop: Stop = None,
        verbose: bool = False,
        default: str = "",
        max_tokens: int = 256,
    ) -> str:
        raise NotImplementedError

    async def classify(
        self,
        *,
        prompt: str,
        choices: tuple[str, ...],
        default: Optional[str] = None,
        verbose: bool = False,
    ) -> tuple[dict[str, float], Optional[str]]:
        raise NotImplementedError

    # Methods below may be deprecated in the future:

    async def relevance(
        self,
        *,
        context: str,
        question: str,
        verbose: bool = False,
        default: Optional[float] = None,
    ) -> float:
        raise NotImplementedError

    async def predict(
        self, *, context: str, default: str = "", verbose: bool = False
    ) -> dict[str, float]:
        raise NotImplementedError



import math

from structlog.stdlib import get_logger
from tqdm import tqdm

from ice.paper import Paper
from ice.paper import Paragraph
from ice.recipe import Recipe
from ice.utils import nsmallest_async

log = get_logger()


def make_compare_paragraphs_prompt(a: Paragraph, b: Paragraph, question: str) -> str:
    return f"""
Which of paragraphs A and B better answers the question "{question}"?

Paragraph A: {a}

Paragraph B: {b}

Question: Which of paragraphs A and B better answers the question '{question}'? Answer with "Paragraph A" or "Paragraph B".

Answer: Paragraph""".strip()


class RankParagraphs(Recipe):
    async def run(
        self, paper: Paper, question: str = "What are the interventions?", n: int = 5
    ) -> list[Paragraph]:
        """
        Rank the paragraphs by how well they answers the question
        using binary search, repeatedly asking the question "Which of
        paragraphs A and B better answer question Q?"
        """

        async def cmp(a: Paragraph, b: Paragraph) -> int:
            progress_bar.update(1)
            answer = (
                await self.agent().complete(
                    prompt=make_compare_paragraphs_prompt(a, b, question),
                    stop=[" ", "\n"],
                    max_tokens=1,
                )
            ).strip()
            if answer == "A":
                return -1
            if answer == "B":
                return 1
            log.warning(f"Unrecognized answer: {answer}")
            return 0

        paragraphs = paper.nonempty_paragraphs()

        num_estimated_comparisons = int(len(paragraphs) * math.log2(n))
        with tqdm(total=num_estimated_comparisons) as progress_bar:
            ranked_paragraphs = await nsmallest_async(
                n, paragraphs, cmp, self.max_concurrency()
            )

        return ranked_paragraphs
